# """
# Implementação de um canal para aplicação de ruido AWGN.

# Autor: Arthur Cadore
# Data: 16-08-2025
# """

import numpy as np
from .datagram import Datagram
from .transmitter import Transmitter
from .plotter import save_figure, create_figure, TimePlot, FrequencyPlot, GaussianNoisePlot

class Noise:
    def __init__(self, snr=15, seed=None, length_multiplier=1, position_factor=0.5):
        r"""
        Implementação de canal para aplicação de ruido $AWGN$, com base em $SNR$.

        Args:
            snr (float): Relação sinal-ruído em decibéis (dB).
            seed (int): Seed do gerador de números aleatórios.
            length_multiplier (float): Multiplicador de comprimento do sinal.
            position_factor (float): Fator de posição do ruído.

        Example: 
            ![pageplot](assets/example_noise_time.svg) 
        """
        self.snr = snr
        self.rng = np.random.default_rng(seed)
        self.length_multiplier = length_multiplier
        self.position_factor = np.clip(position_factor, 0, 1)

    
    def add_noise(self, signal):
        r"""
        Adiciona ruído AWGN $n(t)$ ao sinal de entrada $s(t)$, com base na $\mathrm{SNR}_{dB}$ definida na inicialização. 

        $$
        r(t) = s(t) + n(t), \qquad n(t) \sim \mathcal{N}(0, \sigma^2)
        $$

        Sendo: 
            - $r(t)$: Sinal retornado com ruído AWGN adicionado.
            - $s(t)$: Sinal de entrada sem ruído. 
            - $n(t)$: Ruído adicionado, com distribuição normal $\mathcal{N}(0, \sigma^2)$.

        A variância do ruído $\sigma^2$ é dada por:

        $$
        \sigma^2 = \frac{\mathbb{E}\!\left[ |s(t)|^2 \right]}{10^{\frac{\mathrm{SNR}_{dB}}{10}}}
        $$

        Sendo: 
            - $\sigma^2$: A variância do ruído.
            - $\mathbb{E}\!\left[ |s(t)|^2 \right]$: Potência média do sinal de entrada.
            - $\mathrm{SNR}_{dB}$: Relação sinal-ruído em decibéis (dB).

        Args:
            signal (np.ndarray): Sinal transmitido $s(t)$.

        Returns:
            signal (np.ndarray): Sinal $r(t)$, com ruído AWGN adicionado.

        Example:
            ![pageplot](assets/example_noise_gaussian_snr.svg)
        """

        self.signal_power = np.mean(np.abs(signal) ** 2)
        self.snr_linear = 10 ** (self.snr / 10)
        self.variance = self.signal_power / self.snr_linear

        sig_len = len(signal)
        noise_len = int(sig_len * self.length_multiplier)

        # gera vetor de ruído maior
        noise = self.rng.normal(0, np.sqrt(self.variance), noise_len)

        # calcula posição de inserção do sinal
        start_idx = int((noise_len - sig_len) * self.position_factor)
        end_idx = start_idx + sig_len

        # insere o sinal no ruído
        noisy_signal = noise.copy()
        noisy_signal[start_idx:end_idx] += signal

        return noisy_signal

class NoiseEBN0:
    def __init__(self, ebn0_db=10, fs=128_000, Rb=400, seed=None, length_multiplier=1, position_factor=0.5):
        r"""
        Implementação de canal para aplicação de ruido $AWGN$, com base em $Eb/N_{0}$.

        Args:
            ebn0_db (float): Valor alvo de $Eb/N_{0}$ em $dB$
            fs (int): Taxa de amostragem do sinal em $Hz$.
            Rb (int): Taxa de bits em bits/s.
            seed (int): Seed do gerador de números aleatórios.
            length_multiplier (float): Multiplicador de comprimento do sinal.
            position_factor (float): Fator de posição do ruído.
        
        Example: 
            ![pageplot](assets/example_noise_time.svg)
        """
        self.ebn0_db = ebn0_db
        self.ebn0_lin = 10 ** (ebn0_db / 10)
        self.fs = fs
        self.Rb = Rb
        self.rng = np.random.default_rng(seed)
        self.length_multiplier = length_multiplier
        self.position_factor = np.clip(position_factor, 0, 1)

    def add_noise(self, signal):
        r"""
        Adiciona ruído AWGN $n(t) ao sinal de entrada $s(t), com base na $Eb/N0_{dB}$ definida na inicialização. 

        $$
        r(t) = s(t) + n(t), \qquad n(t) \sim \mathcal{N}(0, \sigma^2)
        $$

        Sendo: 
            - $r(t)$: Sinal retornado com ruído AWGN adicionado.
            - $s(t)$: Sinal de entrada sem ruído. 
            - $n(t)$: Ruído adicionado, com distribuição normal $\mathcal{N}(0, \sigma^2)$.

        
        A variância do ruído $\sigma^2$ é dada por:

        $$
        \sigma^2 = \frac{N_0 \cdot f_s}{2}
        $$

        Sendo: 
            - $\sigma^2$: A variância do ruído.
            - $N_0$: Densidade espectral de ruído.
            - $f_s$: Taxa de amostragem do sinal em $Hz$.

        
        A densidade espectral de ruído $N_0$ é dada por:

        $$
        N_0 = \frac{\mathbb{E}\!\left[ |s(t)|^2 \right]}{R_b \cdot 10^{\frac{Eb/N_0}{10}}}
        $$

        Sendo: 
            - $N_0$: Densidade espectral de ruído.
            - $\mathbb{E}\!\left[ |s(t)|^2 \right]$: Potência média do sinal amostrado.
            - $R_b$: Taxa de bits em bits/s.
            - $Eb/N_0$: Relação $dB$ da energia por bit $E_b$ por densidade espectral de ruído $N_0$ dada na inicialização.

        Args:
            signal (np.ndarray): Sinal transmitido $s(t)$.

        Returns:
            signal (np.ndarray): Sinal recebido $r(t)$, com ruído AWGN adicionado.

        Example:
            ![pageplot](assets/example_noise_gaussian_ebn0.svg)

        <div class="referencia">
          <b>Referência:</b>
          <p>Digital communications / John G. Proakis, Masoud Salehi.—5th ed. (pg. 283)</p>
          <p>https://rwnobrega.page/posts/snr/</p>
        </div>
        """

        self.signal_power = np.mean(np.abs(signal)**2)
        self.bit_energy = self.signal_power / self.Rb

        # densidade espectral de ruído
        self.noise_density = self.bit_energy / self.ebn0_lin
        self.variance = (self.noise_density * self.fs) / 2.0

        sig_len = len(signal)
        noise_len = int(sig_len * self.length_multiplier)

        # vetor de ruído maior
        noise = self.rng.normal(0, np.sqrt(self.variance), noise_len)

        # posição de inserção
        start_idx = int((noise_len - sig_len) * self.position_factor)
        end_idx = start_idx + sig_len

        # insere o sinal
        noisy_signal = noise.copy()
        noisy_signal[start_idx:end_idx] += signal

        return noisy_signal

def check_ebn0(s, s_noisy, add_noise:NoiseEBN0):
    n_est = s_noisy - s
    P = np.mean(s**2)
    Eb = P / add_noise.Rb
    # de sigma^2 -> N0 estimado:
    sigma2_meas = np.var(n_est)
    N0_meas = 2 * sigma2_meas / add_noise.fs
    ebn0_meas_db = 10*np.log10(Eb / N0_meas)
    print("Eb/N0 alvo:", add_noise.ebn0_db, "dB | medido:", ebn0_meas_db, "dB")
    

if __name__ == "__main__":
    datagram = Datagram(pcdnum=1234, numblocks=1)
    transmitter = Transmitter(output_print=False, output_plot=False)
    t, s = transmitter.transmit(datagram)

    # ADIÇÃO DE RUIDO USANDO SNR
    snr_db = 15
    add_noise = Noise(snr=snr_db, seed=0)
    s_noisy = add_noise.add_noise(s)

    fig_gauss, grid_gauss = create_figure(1, 1, figsize=(16, 9))
    GaussianNoisePlot(
        fig_gauss, grid_gauss, (0,0),
        variance=add_noise.variance,
        colors="darkorange",
        legend=f"Ruído AWGN SNR - {snr_db} dB",
        xlim=(-1, 1),
        span=200
    ).plot()
    save_figure(fig_gauss, "example_noise_gaussian_snr.pdf")

    # ADIÇÃO DE RUIDO USANDO EBN0
    eb_n0 = 10
    add_noise = NoiseEBN0(ebn0_db=eb_n0, seed=0)
    s_noisy = add_noise.add_noise(s)
    check_ebn0(s, s_noisy, add_noise)

    fig_gauss, grid_gauss = create_figure(1, 1, figsize=(16, 9))
    GaussianNoisePlot(
        fig_gauss, grid_gauss, (0,0),
        variance=add_noise.variance,
        colors="darkorange",
        legend=f"Ruído AWGN Eb/N0 - {eb_n0} dB",
        xlim=(-1, 1)
    ).plot()
    save_figure(fig_gauss, "example_noise_gaussian_ebn0.pdf")


    fig_time, grid_time = create_figure(2, 1, figsize=(16, 9))

    TimePlot(
        fig_time, grid_time, (0,0),
        t=t,
        signals=[s],
        labels=["$s(t)$"],
        title="Domínio do Tempo - Sem Ruído",
        xlim=(40, 200),
        amp_norm=True,
        colors="darkblue",
        style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()
    
    TimePlot(
        fig_time, grid_time, (1,0),
        t=t,
        signals=[s_noisy],
        labels=["$s(t) + AWGN$"],
        title="Domínio do Tempo - Com Ruído",
        xlim=(40, 200),
        amp_norm=True,
        colors="darkred",
        style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()
    
    fig_time.tight_layout()
    save_figure(fig_time, "example_noise_time.pdf")

    fig_freq, grid_freq = create_figure(2, 1, figsize=(16, 9))

    FrequencyPlot(
        fig_freq, grid_freq, (0,0),
        fs=transmitter.fs,
        signal=s,
        fc=transmitter.fc,
        labels=["$S(f)$"],
        title="Domínio da Frequência - Sem Ruído",
        xlim=(-8, 8),
        colors="darkblue",
        style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()
    
    FrequencyPlot(
        fig_freq, grid_freq, (1,0),
        fs=transmitter.fs,
        signal=s_noisy,
        fc=transmitter.fc,
        labels=["$S(f)$ + AWGN"],
        title="Domínio da Frequência - Com Ruído",
        xlim=(-8, 8),
        colors="darkred",
        style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()
    
    fig_freq.tight_layout()
    save_figure(fig_freq, "example_noise_freq.pdf")


# """
# Implementação do multiplexador. O multiplexador concatena os vetores I e Q de dois canais, conforme o padrão PPT-A3.

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
from .plotter import create_figure, save_figure, BitsPlot

class Multiplexer:
    def __init__(self):
        r"""
        Inicializa o multiplexador no padrão ARGOS-3.
        
        Example: 
            ![pageplot](assets/example_mux.svg)
        """
        pass

    def concatenate(self, SI, SQ, Xn, Yn):
        r"""
        Concatena os vetores $X_n$ e $Y_n$ de entrada, com $S_I$ e $S_Q$, retornando os vetores concatenados $X_n$ e $Y_n$. O processo de multiplexação é dado pela expressão abaixo.

        $$
        \begin{align}
        X_n = S_I \oplus X_n \text{ , } \quad Y_n = S_Q \oplus Y_n
        \end{align}
        $$

        Args:
            SI (np.ndarray): Vetor de entrada $S_I$.
            SQ (np.ndarray): Vetor de entrada $S_Q$.
            Xn (np.ndarray): Vetor de entrada $X_n$.
            Yn (np.ndarray): Vetor de entrada $Y_n$.

        Returns:
            Xn (np.ndarray): Vetor $X_n$ concatenado.
            Yn (np.ndarray): Vetor $Y_n$ concatenado.
        
        Raises:
            AssertionError: Se os vetores I e Q não tiverem o mesmo comprimento em ambos os canais.
        """
        assert len(SI) == len(SQ) and len(Xn) == len(Yn), "Os vetores I e Q devem ter o mesmo comprimento em ambos os canais."

        Xn = np.concatenate((SI, Xn))
        Yn = np.concatenate((SQ, Yn))

        return Xn, Yn

if __name__ == "__main__":

    mux = Multiplexer()

    SI = np.random.randint(0, 2, 15)
    SQ = np.random.randint(0, 2, 15)
    X = np.random.randint(0, 2, 60)
    Y = np.random.randint(0, 2, 60)
    print("SI:", ''.join(str(int(b)) for b in SI))
    print("SQ:", ''.join(str(int(b)) for b in SQ))
    print("X: ", ''.join(str(int(b)) for b in X))
    print("Y: ", ''.join(str(int(b)) for b in Y))

    fig_mux, grid_mux = create_figure(2, 1, figsize=(16, 9))

    BitsPlot(
        fig_mux, grid_mux, (0,0),
        bits_list=[SI, X],
        sections=[("Preambulo $S_I$", len(SI)),
                  ("Canal I $(X_n)$", len(X))],
        colors=["blue", "purple"],
        ylabel="Canal $I$"
    ).plot()
    
    Xn, Yn = mux.concatenate(SI, SQ, X, Y)

    BitsPlot(
        fig_mux, grid_mux, (1,0),
        bits_list=[SQ, Y],
        sections=[("Preambulo $S_Q$", len(SQ)),
                  ("Canal Q $(Y_n)$", len(Y))],
        colors=["blue", "purple"],
        xlabel="Index de Bit", 
        ylabel="Canal $Q$"
    ).plot()

    fig_mux.tight_layout()
    save_figure(fig_mux, "example_mux.pdf")

    print("Xn:", ''.join(str(int(b)) for b in Xn))
    print("Yn:", ''.join(str(int(b)) for b in Yn))



# """
# Implementa uma classe de filtro casado para maximizar a SNR do sinal recebido.

# Autor: Arthur Cadore
# Data: 15-08-2025
# """

import numpy as np
from .plotter import create_figure, save_figure, ImpulseResponsePlot, TimePlot, BitsPlot
from .formatter import Formatter
from .encoder import Encoder

class MatchedFilter:
    def __init__(self, alpha=0.8, fs=128_000, Rb=400, span=6, type="RRC-Inverted", channel=None, bits_per_symbol=1):
        r"""
        Inicializa um filtro casado. O filtro casado é usado para maximizar a SNR do sinal recebido.

        Args:
            alpha (float): Fator de roll-off do filtro casado.
            fs (int): Frequência de amostragem.
            Rb (int): Taxa de bits.
            span (int): Duração do pulso em termos de períodos de bit.
            type (str): Tipo de filtro, atualmente apenas "RRC-Inverted" e "Manchester-Inverted" são suportados.

        Raises:     
            ValueError: Se o tipo de filtro não for suportado.

        Example: 
            ![pageplot](assets/receiver_mf_time.svg) 
        """
        self.alpha = alpha
        self.fs = fs
        self.Rb = Rb
        self.Tb = 1 / Rb
        self.sps = int(fs / Rb)
        self.span = span
        self.channel = channel
        self.bits_per_symbol=bits_per_symbol
        type_map = {
            "rrc-inverted": 0,
            "manchester-inverted": 1
        }

        type = type.lower()
        if type not in type_map:
            raise ValueError("Tipo de filtro inválido. Use 'RRC-inverted' ou 'Manchester-inverted'.")
        
        self.type = type_map[type]

        if self.type == 0:  # RRC
            self.formatter = Formatter(alpha=self.alpha, fs=self.fs, Rb=self.Rb, span=self.span, type="RRC", channel=self.channel, bits_per_symbol=self.bits_per_symbol)
        elif self.type == 1:  # Manchester
            self.formatter = Formatter(alpha=self.alpha, fs=self.fs, Rb=self.Rb, span=self.span, type="Manchester", channel=self.channel, bits_per_symbol=self.bits_per_symbol)
        
        self.g = self.formatter.g
        self.t_rc = self.formatter.t_rc

        # Inverte o pulso
        self.g_inverted = self.inverted_pulse(self.g)

    def inverted_pulse(self, pulse):
        r"""
        Inverte o pulso.

        Args:
            pulse (np.ndarray): Pulso a ser invertido.

        Returns:
            pulse_inverted (np.ndarray): Pulso invertido.

        Example:
            - RRC: ![pageplot](assets/example_mf_impulse.svg)
            - Manchester: ![pageplot](assets/example_mf_impulse_man.svg)

        """
        return pulse[::-1]


    def apply_filter(self, signal):
        r"""
        Aplica o filtro casado com resposta ao impulso $g(-t)$ ao sinal de entrada $s(t)$. O processo de filtragem é dado pela expressão abaixo. 

        $$
            x(t) = s(t) \ast g(-t)
        $$

        Sendo: 
            - $x(t)$: Sinal filtrado.
            - $s(t)$: Sinal de entrada.
            - $g(-t)$: Pulso formatador $RRC$ invertido.

        Args:
            signal (np.ndarray): Sinal de entrada $s(t)$.

        Returns:
            signal_filtered (np.ndarray): Sinal filtrado $x(t)$.
        """
        # convolução completa
        y_full = np.convolve(signal, self.g_inverted, mode='full')

        # atraso do filtro (group delay)
        delay = (len(self.g_inverted) - 1) // 2

        # compensação: extrai a parte alinhada com o sinal original
        start = delay
        end = start + len(signal)
        if end > len(y_full):  # padding de segurança
            y_full = np.pad(y_full, (0, end - len(y_full)), mode='constant')

        signal_filtered = y_full[start:end]

        # normalização segura
        pulse_energy = np.sum(self.g_inverted**2)
        if pulse_energy > 0:
            signal_filtered = signal_filtered / pulse_energy

        return signal_filtered

if __name__ == "__main__":

    bit1 = np.random.randint(0, 2, 10)
    bit2 = bit1

    encoder_1 = Encoder(method="Manchester")
    encoder_2 = Encoder(method="NRZ")

    Yman1 = encoder_1.encode(bit1)
    Yman2 = encoder_2.encode(bit2)

    formatterQ1 = Formatter(alpha=0.8, fs=128_000, Rb=400, span=6, type="RRC", channel="Q", bits_per_symbol=1)
    formatterQ2 = Formatter(alpha=0.8, fs=128_000, Rb=400, span=6, type="Manchester", channel="Q", bits_per_symbol=2)

    dQ1 = formatterQ1.apply_format(Yman1, add_prefix=False)
    dQ2 = formatterQ2.apply_format(Yman2, add_prefix=False)

    filtroQ1 = MatchedFilter(alpha=0.8, fs=128_000, Rb=400, span=12, type="RRC-Inverted", channel="Q", bits_per_symbol=1)
    filtroQ2 = MatchedFilter(alpha=0.8, fs=128_000, Rb=400, span=12, type="Manchester-Inverted", channel="Q", bits_per_symbol=2)

    fig_impulse, grid_impulse = create_figure(1, 1, figsize=(16, 5))
    ImpulseResponsePlot(
        fig_impulse, grid_impulse, (0,0),
        filtroQ1.t_rc, [filtroQ1.g, filtroQ1.g_inverted],
        t_unit="ms",
        colors=["darkorange", "steelblue"],
        label=[r"$g(t)$", r"$g(-t)$"],
        xlabel=r"Tempo ($ms$)",
        ylabel="Amplitude",
        xlim=(-15, 15),
        amp_norm=True
    ).plot()
    fig_impulse.tight_layout()
    save_figure(fig_impulse, "example_mf_impulse.pdf")
    

    fig_impulse, grid_impulse = create_figure(1, 1, figsize=(16, 5))
    ImpulseResponsePlot(
        fig_impulse, grid_impulse, (0,0),
        filtroQ2.t_rc, [filtroQ2.g, filtroQ2.g_inverted],
        t_unit="ms",
        colors=["darkorange", "steelblue"],
        label=[r"$g(t)$", r"$g(-t)$"],
        xlabel=r"Tempo ($ms$)",
        ylabel="Amplitude",
        xlim=(-15, 15),
        amp_norm=True
    ).plot()
    fig_impulse.tight_layout()
    save_figure(fig_impulse, "example_mf_impulse_man.pdf")


    dQ1_filtered = filtroQ1.apply_filter(dQ1)
    dQ2_filtered = filtroQ2.apply_filter(dQ2)

    fig_time, grid_time = create_figure(3, 2, figsize=(16, 9))

    BitsPlot(
        fig_time, grid_time, (0,0),
        bits_list=[bit1],
        sections=[("Bits", len(bit1))],
        title=r"Bits $Q1$",
        colors=["darkblue"],
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    BitsPlot(
        fig_time, grid_time, (0,1),
        bits_list=[bit2],
        sections=[("Bits", len(bit2))],
        title=r"Bits $Q2$",
        colors=["darkblue"],
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    TimePlot(
        fig_time, grid_time, (1,0),
        t= np.arange(len(dQ1)) / formatterQ1.fs,
        signals=[dQ1],
        labels=[r"$d_Q1(t)$"],
        title=r"Canal $Q1$",
        amp_norm=True,
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    TimePlot(
        fig_time, grid_time, (1,1),
        t= np.arange(len(dQ2)) / formatterQ2.fs,
        signals=[dQ2],
        labels=[r"$d_Q2(t)$"],
        title=r"Canal $Q2$",
        # xlim=(40, 140),
        amp_norm=True,
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    TimePlot(
        fig_time, grid_time, (2,0),
        t= np.arange(len(dQ1_filtered)) / formatterQ1.fs,
        signals=[dQ1_filtered],
        labels=[r"$d_Q1(t)$"],
        title=r"Canal $Q1$",
        # xlim=(40, 140),
        amp_norm=True,
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    TimePlot(
        fig_time, grid_time, (2,1),
        t= np.arange(len(dQ2_filtered)) / formatterQ2.fs,
        signals=[dQ2_filtered],
        labels=[r"$d_Q2(t)$"],
        title=r"Canal $Q2$",
        # xlim=(40, 140),
        amp_norm=True,
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    fig_time.tight_layout()
    save_figure(fig_time, "example_mf_time.pdf")


# """
# Implementação de um detector de portadora para recepção PTT-A3.

# Autor: Arthur Cadore
# Data: 07-09-2025
# """

import numpy as np
from .plotter import create_figure, save_figure, PowerMatrixPlot, MatrixSquarePlot, DetectionFrequencyPlot, PowerMatrix3DPlot
from .datagram import Datagram
from .transmitter import Transmitter
from .receiver import Receiver
from .channel import Channel


class CarrierDetector:
    def __init__(self, fs: float = 128_000, seg_ms: float = 10.0, threshold: float = -10, freq_window: tuple[float, float] = (0000, 10000), bandwidth: float = 1600, history: int = 4):
        """
        Inicializa um detector de portadora, utilizado para detectar possíveis portadoras no sinal recebido.

        Args:
            fs (float): Frequência de amostragem [Hz]
            seg_ms (float): Duração de cada segmento [ms]
            threshold (float): Limiar de potência para detecção
            freq_window (tuple[float, float]): Intervalo de frequências (`f_min`, `f_max`).Frequências fora deste intervalo serão descartadas.
        
        Raises:
            ValueError: Se a frequência de amostragem for menor ou igual a zero.
            ValueError: Se o comprimento de cada segmento for menor ou igual a zero.

        Example: 
            - Segmentos de tempo: ![pageplot](assets/example_detector_freq.svg)
            - Diagrama Waterfall: ![pageplot](assets/example_detector_waterfall.svg)
            - Diagrama Waterfall de detecção: ![pageplot](assets/example_detector_waterfall_detection.svg)
            - Diagrama Waterfall de decisão: ![pageplot](assets/example_detector_waterfall_decision.svg)


        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-2097-CNES (Seção 3.3)
        </div>
        """
        if fs <= 0:
            raise ValueError("A frequência de amostragem deve ser maior que zero.")
        if seg_ms <= 0:
            raise ValueError("O comprimento de cada segmento deve ser maior que zero.")

        self.fs = fs
        self.ts = 1 / self.fs
        self.seg_s = seg_ms / 1000.0
        self.N = int(self.fs * self.seg_s)
        self.threshold = threshold
        self.freq_window = freq_window
        self.bandwidth = bandwidth
        self.history = history

        # Valores fixos de espectro
        self.delta_f = self.fs / self.N
        self.span = self.delta_f / 2

        # Quantos bins de FFT correspondem à largura de banda
        self.bandwidth_bins = int(self.bandwidth / self.delta_f)


        self.power_matrix = None
        self.detected_matrix = None
        self.decision_matrix = None
        
    def segment_signal(self, signal: np.ndarray) -> list[np.ndarray]:
        r"""
        Divide o sinal recebido em segmentos de tempo $x_n[m]$, cada segmento com `seg_ms` de duração, conforme a expressão abaixo. 

        $$
        x_n[m] = s(t_{n} + mT_s)
        $$

        Sendo: 
            - $x_n[m]$ : Segmento de tempo $n$.
            - $s(t)$ : Sinal recebido.
            - $T_s$ : Período de amostragem.
            - $m$ : Número do segmento.
            - $t_n$ : Instante de início do segmento $n$.

        Args:
            signal (np.ndarray): sinal recebido

        Returns:
            list[np.ndarray]: lista de segmentos de tempo
        """
        segments = []
        total_samples = len(signal)

        for start in range(0, total_samples, self.N):
            end = min(start + self.N, total_samples)
            segments.append(signal[start:end])

        return segments


    def analyze_signal(self, signal: np.ndarray):
        r"""
        Calcula a FFT de cada segmento $x_n[m]$, usando a expressão abaixo. 
        
        $$
            X_n[k] = \sum_{m=0}^{N-1} x_n[m]\, e^{-j2\pi km/N} 
        $$

        Sendo: 
            - $X_n[k]$ : Transformada de Fourier do segmento $n$.
            - $x_n[m]$ : Segmento de tempo $n$.
            - $N$ : Número de amostras do segmento.
            - $k$ : Número da transformada de Fourier.
            - $m$ : Número da amostra.
            - $T_s$ : Período de amostragem.
            - $e^{-j2\pi km/N}$ : Exponencial complexa.

        Em seguida, calcula a potência espectral $P_n[k]$ em $dB$, e divide pelo número de amostras $N$ contidas no segmento para normalização.

        $$
            P_n[k] = \frac{|X_n[k]|^2}{N}
        $$

        Sendo: 
            - $P_n[k]$ : Potência espectral do segmento $n$, normalizada em $dB$.
            - $X_n[k]$ : Transformada de Fourier do segmento $n$.
            - $N$ : Número de amostras do segmento.

        Args:
            segment (np.ndarray): segmento de tempo

        Returns:
            freqs (tuple[np.ndarray,np.ndarray]): tupla com as frequências e a potência espectral em $dB$

        Example: 
            Matriz de Potência 2D: ![pageplot](assets/example_detector_waterfall.svg)
            Matriz de Potência 3D: ![pageplot](assets/example_detector_waterfall_3d.svg)
        """

        segments = self.segment_signal(signal)
        n_segments = len(segments)
        n_freqs = self.N // 2 + 1

        self.power_matrix = np.zeros((n_segments, n_freqs))

        for i, seg in enumerate(segments):
            X = np.fft.rfft(seg, n=self.N)
            P_bin = (np.abs(X) ** 2) / len(seg)  # normaliza pelo tamanho real do segmento
            P_db = 10.0 * np.log10(P_bin + 1e-12)  # evita log(0)
            self.power_matrix[i, :] = P_db


    def detect(self, s: np.ndarray):
        r"""
        Detecta possíveis portadoras no sinal, comparando $P_n[k]$ com o limiar $P_t$, para cada índice $k$ da FFT.

        $$
            f_n[k] =
            \begin{cases}
            \dfrac{k}{N} \cdot f_s, & \text{se } P_n[k] > P_t\\
            \text{não detectada}, & \text{se } P_n[k] \leq P_t
            \end{cases}
        $$

        Sendo: 
            - $f_n[k]$ : frequência detectada no segmento $n$.
            - $P_n[k]$ : potência espectral do segmento $n$.
            - $P_t$ : limiar de potência.
            - $N$ : número de amostras do segmento.
            - $f_s$ : frequência de amostragem.
            - $k$ : índice da FFT.
            - `não detectada`: Frequência ignorada no processo de detecção.  

        Args:
            s (np.ndarray): sinal recebido

        Returns:
            results (list[tuple[np.ndarray, list[float]]]): lista de tuplas com os segmentos e as frequências detectadas

        Example: 
            ![pageplot](assets/example_detector_waterfall_detection.svg)
        """
        # Calcula matriz de potência FFT
        self.analyze_signal(s)

        n_segments, n_freqs = self.power_matrix.shape
        self.detected_matrix = np.zeros((n_segments, n_freqs), dtype=int)

        # Frequências reais dos bins da FFT
        freqs = np.fft.rfftfreq(self.N, d=self.ts)

        for i in range(n_segments):
            P_db = self.power_matrix[i, :]

            # Máscara de detecção pelo limiar
            mask = P_db > self.threshold

            # Restringe à janela de frequências
            if self.freq_window is not None:
                fmin, fmax = self.freq_window
                mask &= (freqs >= fmin) & (freqs <= fmax)

            detected_bins = np.where(mask)[0]

            for k in detected_bins:
                if i >= self.history:
                    # confirma somente se todos os últimos 'history' forem exatamente 1
                    past_values = self.detected_matrix[i-self.history:i, k]
                    if np.all(past_values == 1):

                        # frequência confirmada, vai pra demodulação
                        self.detected_matrix[i, k] = 2
                    else:
                        # frequência detectada, mas não confirmada
                        self.detected_matrix[i, k] = 1
                else:
                    # apenas detectado, sem histórico
                    self.detected_matrix[i, k] = 1

        self.decision()

    def decision(self):
        """
        Retorna apenas frequências que foram detectadas em dois segmentos consecutivos. A tolerância é dada pela resolução espectral da FFT, $\Delta f$, conforme a expressão abaixo. 

        $$
            \Delta f = \dfrac{f_s}{N}
        $$

        Sendo: 
            - $\Delta f$ : resolução espectral da FFT.
            - $f_s$ : frequência de amostragem.
            - $N$ : número de amostras do segmento.

        Args:
            results (list[tuple[np.ndarray, list[float]]]): saída de self.detect()
            confirmed_freqs (list[float]): lista de frequências confirmadas como portadora

        Returns:
            confirmed_freqs (list[float]): lista de frequências confirmadas como portadora

        Example: 
            ![pageplot](assets/example_detector_waterfall_decision.svg)
        """

        self.decision_matrix = np.copy(self.detected_matrix)
        n_segments, n_freqs = self.detected_matrix.shape

        # matriz auxiliar para controlar spans existentes
        span_matrix = np.zeros_like(self.detected_matrix, dtype=bool)

        runs = []

        half_span = (self.bandwidth_bins - 1) // 2  # calcula metade do span para aplicar acima e abaixo do centro

        for i in range(n_segments):
            for k in range(n_freqs):
                # só processa centros detectados (2) que não estão dentro de um span existente
                if self.detected_matrix[i, k] != 2 or span_matrix[i, k]:
                    continue

                center_k = k
                s = i + 1  # começa no próximo segmento
                zero_count = 0
                start_s = s

                # aplica o 4 e o span no primeiro segmento após o 2
                lower = max(center_k - half_span, 0)
                upper = min(center_k + half_span, n_freqs - 1)
                self.decision_matrix[s, lower:upper + 1] = np.where(
                    np.arange(lower, upper + 1) == center_k,
                    4,  # centro
                    3   # span
                )
                span_matrix[s, lower:upper + 1] = True

                s += 1  # avança para continuar o loop de extensão

                # agora continua preenchendo a sequência enquanto houver atividade
                while s < n_segments and zero_count < 2:
                    neighbors = [center_k]
                    if center_k > 0:
                        neighbors.append(center_k - 1)
                    if center_k < n_freqs - 1:
                        neighbors.append(center_k + 1)

                    found_activity = False
                    for look_ahead in range(0, 3):
                        idx = s + look_ahead
                        if idx >= n_segments:
                            break
                        if any(self.detected_matrix[idx, nb] in (1, 2) for nb in neighbors):
                            found_activity = True
                            break

                    # aplica o span no segmento atual
                    self.decision_matrix[s, lower:upper + 1] = np.where(
                        np.arange(lower, upper + 1) == center_k,
                        4,
                        3
                    )
                    span_matrix[s, lower:upper + 1] = True

                    if found_activity:
                        zero_count = 0
                    else:
                        zero_count += 1

                    s += 1

                runs.append((start_s, s - 1, center_k))

    def return_channels(self):
        """
        Varre a decision_matrix e retorna as frequências confirmadas.
        com o início e fim do segmento onde a portadora foi demodulada.

        Returns:
            channels (list[tuple[float, int, int]]): Lista de tuplas (freq_Hz, start_segment, end_segment)
        """
        if not hasattr(self, 'decision_matrix'):
            raise ValueError("A decision_matrix ainda não foi criada. Execute self.decision() antes.")

        n_segments, n_freqs = self.decision_matrix.shape
        visited = np.zeros_like(self.decision_matrix, dtype=bool)
        channels = []

        # Frequências reais dos bins da FFT
        freqs = np.fft.rfftfreq(self.N, d=self.ts)

        for i in range(n_segments):
            for k in range(n_freqs):
                # só processa centros não visitados
                if self.decision_matrix[i, k] != 4 or visited[i, k]:
                    continue

                start_segment = i
                s = i
                # percorre os segmentos enquanto houver 4 no centro
                while s < n_segments and self.decision_matrix[s, k] == 4:
                    visited[s, k] = True
                    s += 1
                end_segment = s - 1

                channels.append((freqs[k], start_segment, end_segment))

        return channels

if __name__ == "__main__":

    fs = 128_000
    Rb = 400
    
    fc1 = np.random.randint(10, 30)*100
    fc2 = fc1 + 2500
    fc3 = fc2 + 2500
    
    datagram1 = Datagram(pcdnum=1234, numblocks=1, seed=11)
    datagram2 = Datagram(pcdnum=1234, numblocks=4, seed=11)
    datagram3 = Datagram(pcdnum=1234, numblocks=8, seed=11)

    print("ut1: ", ''.join(str(b) for b in datagram1.streambits))
    print("ut2: ", ''.join(str(b) for b in datagram2.streambits))
    print("ut3: ", ''.join(str(b) for b in datagram3.streambits))

    transmitter1 = Transmitter(fc=fc1, fs=fs, Rb=Rb, output_print=False, output_plot=False, carrier_length=0.08)
    transmitter2 = Transmitter(fc=fc2, fs=fs, Rb=Rb, output_print=False, output_plot=False, carrier_length=0.08)
    transmitter3 = Transmitter(fc=fc3, fs=fs, Rb=Rb, output_print=False, output_plot=False, carrier_length=0.08)
    transmitter4 = Transmitter(fc=fc1, fs=fs, Rb=Rb, output_print=False, output_plot=False, carrier_length=0.08)

    print("Gerando vetores de transmissão")
    t1, s1 = transmitter1.transmit(datagram1)
    t2, s2 = transmitter2.transmit(datagram2)
    t3, s3 = transmitter3.transmit(datagram3)
    t4, s4 = transmitter4.transmit(datagram1)

    # Canal
    channel = Channel(fs=fs, duration=1, noise_mode="ebn0", noise_db=20, seed=11)

    # gera distâncias aleatórias. 
    p1 = np.random.choice(np.arange(0, 0.21, 0.1))
    p2 = np.random.choice(np.arange(0, 1.01, 0.1))  
    p3 = np.random.choice(np.arange(0, 1.01, 0.1))  
    p4 = p1 + 0.6

    print("Adicionando sinais ao canal")
    s1 = channel.add_signal(s1, position_factor=p1)
    s2 = channel.add_signal(s2, position_factor=p2)
    s3 = channel.add_signal(s3, position_factor=p3)
    s4 = channel.add_signal(s4, position_factor=p4)

    # Adiciona ruído
    channel.add_noise()
    st = channel.channel


    # Detecção de portadora
    threshold = -15
    detector = CarrierDetector(fs=fs, seg_ms=10, threshold=threshold) 
    detector.detect(st.copy())

    # Heatmap da potência
    fig, grid = create_figure(1, 1, figsize=(16, 9))
    PowerMatrixPlot(fig, grid, 0,
                detector.power_matrix,
                fs=detector.fs, N=detector.N).plot()
    save_figure(fig, "example_detector_waterfall.pdf")

    fig, grid = create_figure(1, 1, figsize=(12, 12))
    PowerMatrix3DPlot(fig, grid, 0,
                      detector.power_matrix,
                      fs=detector.fs,
                      N=detector.N,
                      freq_window=detector.freq_window,
                      threshold=detector.threshold,
                      elev=2, azim=-10
    ).plot()
    
    save_figure(fig, "example_detector_waterfall_3d.pdf")

    # Heatmap da detecção
    fig, grid = create_figure(1, 1)
    MatrixSquarePlot(fig, grid, 0,
                 detector.detected_matrix,
                 fs=detector.fs, 
                 legend_list=["Detectada", "Confirmada"],
                 N=detector.N).plot()

    save_figure(fig, "example_detector_waterfall_detection.pdf")

    # Heatmap da decisão
    fig, grid = create_figure(1, 1)
    MatrixSquarePlot(fig, grid, 0,
                 detector.decision_matrix,
                 fs=detector.fs, 
                 legend_list=["Detectada", "Confirmada", "Span", "Demodulação"],
                 N=detector.N).plot()
    save_figure(fig, "example_detector_waterfall_decision.pdf")

    # plota o espectro do sinal no segmento desejado
    seg_index = 1
    fig, grid = create_figure(2, 1)
    DetectionFrequencyPlot(fig, grid, 0, 
              fs=fs, 
              signal=detector.power_matrix[seg_index, :], 
              threshold=detector.threshold, 
              xlim=(0, 10),
              title="Detecção de portadora de $s(t)$ - Segmento %d" % seg_index,
              labels=["$S(f)$"],
              colors="darkred",
              freqs_detected=detector.detected_matrix[seg_index, :]
    ).plot()
    DetectionFrequencyPlot(fig, grid, 1, 
              fs=fs, 
              signal=detector.power_matrix[seg_index+1, :], 
              threshold=detector.threshold, 
              xlim=(0, 10),
              title="Detecção de portadora de $s(t)$ - Segmento %d" % (seg_index+1),
              labels=["$S(f)$"],
              colors="darkred",
              freqs_detected=detector.detected_matrix[seg_index+1, :]
    ).plot()
    save_figure(fig, "example_detector_freq.pdf")

    # Recepção: 
    channels = detector.return_channels()

    print("Frequências confirmadas (Hz) com início e fim de segmentos:")
    for f, start, end in channels:
        print(f"Frequência {f:.1f} Hz: segmento {start} -> {end}")
    
    for idx, (freq, start_seg, end_seg) in enumerate(channels, start=1):
        print(f"\n ==============================================")
        print(f"\n ==== RECEPÇÃO DE s(t) COM f_c = {freq:.1f} Hz ==== \n")
    
        # Seleciona apenas os segmentos necessários e concatena
        first_segment = int((start_seg - 5) * detector.fs * detector.seg_s)
        last_segment = int(end_seg * detector.fs * detector.seg_s)
        selected_signal = st[first_segment:last_segment]
    
        # Instancia o receptor
        receiver = Receiver(fc=freq, fs=detector.fs, Rb=Rb, output_print=True, output_plot=True)
        datagramRX, success = receiver.receive(selected_signal)
    
        if not success:
            bitsRX = datagramRX
            print("Decodificação incorreta: ")
            print("Bits RX: ", ''.join(str(b) for b in bitsRX))



# """
# Implementa um formatador de pulso para transmissão de sinais digitais. 

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
from .plotter import ImpulseResponsePlot, TimePlot, SymbolsPlot, create_figure, save_figure
from .encoder import Encoder

class Formatter:
    def __init__(self, alpha=0.8, fs=128_000, Rb=400, span=6, type="RRC", prefix_duration=0.082, channel=None, bits_per_symbol=1):
        r"""
        Inicializa um formatador, utilizado preparar os símbolos para modulação.

        Args:
            alpha (float): Fator de roll-off do pulso RRC.
            fs (int): Frequência de amostragem.
            Rb (int): Taxa de bits.
            span (int): Duração do pulso em termos de períodos de bit.
            type (str): Tipo de pulso, atualmente apenas $RRC$ é suportado.
            prefix_duration (int): Duração da portadora pura no inicio do vetor
            channel (str): Canal a ser formatado, apenas $I$ e $Q$ são suportados.
            bits_per_symbol (int): Número de bits por símbolo.

        Raises:
            ValueError: Se o tipo de pulso não for suportado.
            ValueError: Se o canal não for suportado.

        Example: 
            ![pageplot](assets/example_formatter_time.svg)

        <div class="referencia">
        <b>Referência:</b><br>
        EEL7062 – Princípios de Sistemas de Comunicação, Richard Demo Souza (Pg. 55)
        </div>
        """

        if channel not in ["I", "Q"]:
            raise ValueError("Canal inválido. Use 'I' ou 'Q'.")
        
        self.channel = channel
        self.prefix_duration = prefix_duration  
        self.alpha = alpha
        self.bits_per_symbol = bits_per_symbol
        self.fs = fs
        self.Rb = Rb * bits_per_symbol
        self.Tb = 1 / self.Rb
        self.sps = int(fs / self.Rb)
        self.span = span
        self.t_rc = np.linspace(-span * self.Tb, span * self.Tb, span * self.sps * 2)

        type_map = {
            "rrc": 0,
            "manchester": 1
        }


        type = type.lower()
        if type not in type_map:
            raise ValueError("Tipo de pulso inválido. Use 'RRC' ou 'Manchester'.")
        
        self.type = type_map[type]

        if self.type == 0:  # RRC
            self.g = self.rrc_pulse()
        elif self.type == 1:  # Manchester
            self.g, self.g_left, self.g_right = self.manchester_pulse()

    def rrc_pulse(self, shift=0.0):
        r"""
        Gera o pulso Root Raised Cosine ($RRC$). O pulso $RRC$ no dominio do tempo é definido pela expressão abaixo.

        $$
            g(t) = \frac{(1 - \alpha) sinc((1- \alpha) t / T_b) + \alpha (4/\pi) \cos(\pi (1 + \alpha) t / T_b)}{1 - (4 \alpha t / T_b)^2}
        $$

        Sendo: 
            - $g(t)$: Pulso formatador $RRC$ no dominio do tempo.
            - $\alpha$: Fator de roll-off do pulso.
            - $T_b$: Período de bit.
            - $t$: Vetor de tempo.

        Args: 
            shift (float): Deslocamento no tempo.

        Returns:
           rc (np.ndarray): Pulso RRC normalizado.

        Example: 
            ![pageplot](assets/example_formatter_impulse.svg)
        """
        self.t_rc = np.array(self.t_rc, dtype=float)
        # aplica deslocamento no tempo
        t_shifted = self.t_rc - shift

        rc = np.zeros_like(t_shifted)
        for i, ti in enumerate(t_shifted):
            if np.isclose(ti, 0.0):
                rc[i] = 1.0 + self.alpha * (4/np.pi - 1)
            elif self.alpha != 0 and np.isclose(np.abs(ti), self.Tb/(4*self.alpha)):
                rc[i] = (self.alpha/np.sqrt(2)) * (
                    (1 + 2/np.pi) * np.sin(np.pi/(4*self.alpha)) +
                    (1 - 2/np.pi) * np.cos(np.pi/(4*self.alpha))
                )
            else:
                num = np.sin(np.pi * ti * (1 - self.alpha) / self.Tb) + \
                      4 * self.alpha * (ti / self.Tb) * np.cos(np.pi * ti * (1 + self.alpha) / self.Tb)
                den = np.pi * ti * (1 - (4 * self.alpha * ti / self.Tb) ** 2) / self.Tb
                rc[i] = num / den

        # Normaliza energia para 1
        rc = rc / np.sqrt(np.sum(rc**2))
        return rc
    
    def manchester_pulse(self):
        r"""
        Pulso Manchester como diferença de dois RRC simetricamente deslocados. 

        $$
            g_{MAN}(t) = g_{RRC}(t + T_b/2) - g_{RRC}(t - T_b/2)
        $$

        Sendo: 
            - $g_{MAN}(t)$: Pulso formatador Manchester no dominio do tempo.
            - $g_{RRC}(t)$: Pulso formatador $RRC$ no dominio do tempo.
            - $T_b$: Período de bit.
            - $t$: Vetor de tempo.

        Example: 
            ![pageplot](assets/example_formatter_impulse_man.svg)
        """
        g_left = -self.rrc_pulse(shift=self.Tb/2)
        g_right = +self.rrc_pulse(shift=-self.Tb/2)
        g = g_left + g_right

        # Normaliza energia para 1
        g = g / np.sqrt(np.sum(g**2))

        return g, g_left, g_right

    def apply_format(self, symbols, add_prefix=True):
        r"""
        Formata os símbolos de entrada usando o pulso inicializado. O processo de formatação é dado por: 

        $$
           d(t) = \sum_{n} x[n] \cdot g(t - nT_b)
        $$

        Sendo: 
            - $d(t)$: Sinal formatado de saída.
            - $x$: Vetor de símbolos de entrada.
            - $g(t)$: Pulso formatador.
            - $n$: Indice de bit.
            - $T_b$: Período de bit.

        Args:
            symbols (np.ndarray): Vetor de símbolos a serem formatados.
        
        Returns:
            out_symbols (np.ndarray): Vetor formatado com o pulso aplicado.
        """

        # adiciona prefixo
        if add_prefix:
            symbols = self.add_prefix(symbols)

        pulse = self.g
        # samples per symbol agora é por bit dividido pelo número de bits por símbolo
        sps = int(self.fs / (self.Rb / self.bits_per_symbol))

        upsampled = np.zeros(len(symbols) * sps)
        upsampled[::sps] = symbols
        out_sys = np.convolve(upsampled, pulse, mode='same')

        return out_sys

    def add_prefix(self, symbols):
        """
        Adiciona um prefixo de portadora pura no inicio do sinal. Para o canal $I$, adiciona um vetor de simbolos $+1$, para o canal $Q$, adiciona um vetor de simbolos $0$, com duração de `prefix_duration`, pois ao aplicar o modulador `IQ` temos uma portadora pura no inicio do sinal, conforme a expressão abaixo: 

        $$
            s(t) = 1(t) \cdot \cos(2\pi f_c t) - 0(t) \cdot \sin(2\pi f_c t) \mapsto s(t) = \cos(2\pi f_c t)
        $$

        Sendo: 
            - $s(t)$: Sinal modulado.
            - $1(t)$ e $0(t)$: Prefixo de portadora pura.
            - $f_c$: Frequência da portadora.
            - $t$: Vetor de tempo.
        
        Args:
            symbols (np.ndarray): Vetor de símbolos a serem formatados.
        
        Returns:
            symbols (np.ndarray): Vetor de símbolos com prefixo adicionado.
        """
        if self.channel == "I":
            carrier = np.ones(int(self.prefix_duration * self.Rb / self.bits_per_symbol))
        elif self.channel == "Q":
            carrier = np.zeros(int(self.prefix_duration * self.Rb / self.bits_per_symbol))

        symbols = np.concatenate([carrier, symbols])
        return symbols


if __name__ == "__main__":

    bitN = np.random.randint(0, 2, 20)
    bitM = np.random.randint(0, 2, 20)

    encoder_I = Encoder(method="NRZ")
    encoded_Q = Encoder(method="NRZ")

    symbols_I = encoder_I.encode(bitN)
    symbols_Q = encoded_Q.encode(bitM)
    
    formatterI = Formatter(alpha=0.8, fs=128_000, Rb=400, span=12, type="RRC", channel="I", bits_per_symbol=1, prefix_duration=0)
    formatterQ = Formatter(alpha=0.8, fs=128_000, Rb=400, span=12, type="Manchester", channel="Q", bits_per_symbol=2, prefix_duration=0)
    
    dI1 = formatterI.apply_format(symbols_I)
    dQ1 = formatterQ.apply_format(symbols_Q)
    
    print("Xn:",  ' '.join(f"{x:+d}" for x in symbols_I[:10]))
    print("Yn:",  ' '.join(f"{y:+d}" for y in symbols_Q[:10]))
    print("dI:", ''.join(str(b) for b in dI1[:5]))
    print("dQ:", ''.join(str(b) for b in dQ1[:5]))

    # Plotando a resposta ao impulso
    fig_impulse, grid_impulse = create_figure(1, 1, figsize=(16, 5))

    ImpulseResponsePlot(
        fig_impulse, grid_impulse, (0, 0),
        formatterI.t_rc, formatterI.g,
        t_unit="ms",
        colors="darkorange",
        label=r"$g(t)$", 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", 
        xlim=(-15, 15), 
        amp_norm=True
    ).plot()

    fig_impulse.tight_layout()
    save_figure(fig_impulse, "example_formatter_impulse.pdf")


    # Plotando a resposta ao impulso (man)
    fig_impulse_man, grid_impulse_man = create_figure(2, 1, figsize=(16, 9))

    ImpulseResponsePlot(
        fig_impulse_man, grid_impulse_man, (0, 0),
        formatterQ.t_rc, 
        [formatterQ.g_left, formatterQ.g_right],
        t_unit="ms",
        colors=["darkorange", "navy"],
        label=[r"$g_{L}(t)$", r"$g_{R}(t)$"], 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", 
        xlim=(-15, 15), 
    ).plot()

    ImpulseResponsePlot(
        fig_impulse_man, grid_impulse_man, (1, 0),
        formatterQ.t_rc, formatterQ.g,
        t_unit="ms",
        colors="darkorange",
        label=r"$g(t)$", 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", 
        xlim=(-15, 15), 
        amp_norm=True
    ).plot()

    fig_impulse_man.tight_layout()
    save_figure(fig_impulse_man, "example_formatter_impulse_man.pdf")

    # Plotando os sinais formatados
    fig_format, grid_format = create_figure(3, 2, figsize=(16, 12))

    ImpulseResponsePlot(
        fig_format, grid_format, (0,0),
        formatterI.t_rc, formatterI.g,
        t_unit="ms",
        colors="darkorange",
        label=r"$g(t)$", 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", 
        xlim=(-15, 15), 
        amp_norm=True
    ).plot()

    ImpulseResponsePlot(
        fig_format, grid_format, (0,1),
        formatterQ.t_rc, formatterQ.g,
        t_unit="ms",
        colors="darkorange",
        label=r"$g(t)$", 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", 
        xlim=(-15, 15), 
        amp_norm=True
    ).plot()

    SymbolsPlot(
        fig_format, grid_format, (1,0),
        symbols_list=[symbols_I],
        sections=[("$X_n$", len(symbols_I))],
        colors=["darkgreen"],
        xlabel="Index de Bit", 
        ylabel="$X_n$", 
    ).plot()

    SymbolsPlot(
        fig_format, grid_format, (1,1),
        symbols_list=[symbols_Q],
        sections=[("$Y_n$", len(symbols_Q))],
        colors=["darkblue"],
        xlabel="Index de Bit", 
        ylabel="$Y_n$", 
    ).plot()
        
    TimePlot(
        fig_format, grid_format, (2,0),
        t= np.arange(len(dI1)) / formatterI.fs,
        signals=[dI1],
        labels=[r"$d_I(t)$"],
        title=r"Canal $I$",
        amp_norm=True,
        colors="darkgreen",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    TimePlot(
        fig_format, grid_format, (2,1),
        t= np.arange(len(dQ1)) / formatterQ.fs,
        signals=[dQ1],
        labels=[r"$d_Q(t)$"],
        title=r"Canal $Q$",
        amp_norm=True,
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    fig_format.tight_layout()
    save_figure(fig_format, "example_formatter_time.pdf")


# """
# Implementação das operações de plot

# Autor: Arthur Cadore
# Data: 16-08-2025
# """

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import scienceplots 
import os
from typing import Optional, List, Union, Tuple, Dict, Any
from collections import defaultdict
from matplotlib.lines import Line2D
from mpl_toolkits.mplot3d import Axes3D 
from scipy.ndimage import gaussian_filter
from matplotlib.ticker import FuncFormatter, MultipleLocator
from scipy.signal import freqz

# Parâmetros gerais de plotagem
mpl.rcParams["pdf.fonttype"] = 42
mpl.rcParams["ps.fonttype"] = 42
mpl.rcParams["svg.fonttype"] = "none"
mpl.rcParams["savefig.transparent"] = True

# Estilização science para os plots
plt.style.use("science")

# Cores e estilos
mpl.rcParams["text.color"] = "black"
mpl.rcParams["axes.labelcolor"] = "black"
mpl.rcParams["xtick.color"] = "black"
mpl.rcParams["ytick.color"] = "black"
plt.rcParams["figure.figsize"] = (16, 9)

# Fontes e legendas
plt.rc("font", size=16)
plt.rc("axes", titlesize=22, labelsize=22)
plt.rc("xtick", labelsize=16)
plt.rc("ytick", labelsize=16)
plt.rc("legend", fontsize=12, frameon=True)
plt.rc("figure", titlesize=22)


def mag2db(signal: np.ndarray) -> np.ndarray:
    r"""
    Converte a magnitude do sinal para escala logarítmica ($dB$). O processo de conversão é dado pela expressão abaixo.

    $$
     dB(x) = 20 \log_{10}\left(\frac{|x|}{x_{peak} + 10^{-12}}\right)
    $$

    Sendo:
        - $x$: Sinal a ser convertido para $dB$.
        - $x_{peak}$: Pico de maior magnitude do sinal.
        - $10^{-12}$: Constante para evitar divisão por zero.
    
    Args:
        signal: Array com os dados do sinal
        
    Returns:
        Array com o sinal convertido para $dB$
    """
    mag = np.abs(signal)
    peak = np.max(mag) if np.max(mag) != 0 else 1.0
    mag = mag / peak
    return 20 * np.log10(mag + 1e-12)

def create_figure(rows: int, cols: int, figsize: Tuple[int, int] = (16, 9)) -> Tuple[plt.Figure, gridspec.GridSpec]:
    r"""
    Cria uma figura com `GridSpec`, retornando o objeto `fig` e `grid` para desenhar os plots.
    
    Args:
        rows (int): Número de linhas do GridSpec
        cols (int): Número de colunas do GridSpec
        figsize (Tuple[int, int]): Tamanho da figura
        
    Returns:
        Tuple[plt.Figure, gridspec.GridSpec]: Tupla com a figura e o GridSpec
    """
    fig = plt.figure(figsize=figsize)
    grid = gridspec.GridSpec(rows, cols, figure=fig)
    return fig, grid

def save_figure(fig: plt.Figure, filename: str, out_dir: str = "../../out") -> None:
    r"""
    Salva a figura em `<out_dir>/<filename>` a partir do diretório raiz do script. 
    
    Args:
        fig (plt.Figure): Objeto `Figure` do matplotlib
        filename (str): Nome do arquivo de saída
        out_dir (str): Diretório de saída
    
    Raises:
        ValueError: Se o diretório de saída for inválido
    """
    script_dir = os.path.dirname(os.path.abspath(__file__))
    out_dir = os.path.abspath(os.path.join(script_dir, out_dir))
    os.makedirs(out_dir, exist_ok=True)
    save_path = os.path.join(out_dir, filename)
    fig.tight_layout()
    fig.savefig(save_path, bbox_inches="tight", transparent=True)
    plt.close(fig)

class BasePlot:
    r"""
    Classe base para plotagem de gráficos, implementando funcionalidades comuns a todos os plots.
    
    Args:
        ax (plt.Axes): Objeto `Axes` do matplotlib. 
        title (str): Título do plot. 
        labels (Optional[List[str]]): Lista de rótulos para os eixos. 
        xlim (Optional[Tuple[float, float]]): Limites do eixo x `x = [xlim[0], xlim[1]]`. 
        ylim (Optional[Tuple[float, float]]): Limites do eixo y `y = [ylim[0], ylim[1]]`. 
        colors (Optional[Union[str, List[str]]]): Cores do plot. 
        style (Optional[Dict[str, Any]]): Estilo do plot.
    """
    def __init__(self,
                 ax: plt.Axes,
                 title: str = "",
                 labels: Optional[List[str]] = None,
                 xlim: Optional[Tuple[float, float]] = None,
                 ylim: Optional[Tuple[float, float]] = None,
                 colors: Optional[Union[str, List[str]]] = None,
                 style: Optional[Dict[str, Any]] = None) -> None:
        self.ax = ax
        self.title = title
        self.labels = labels
        self.xlim = xlim
        self.ylim = ylim
        self.colors = colors
        self.style = style or {}

    # Aplica estilos gerais ao eixo
    def apply_ax_style(self) -> None:
        grid_kwargs = self.style.get("grid", {"alpha": 0.6, "linestyle": "--", "linewidth": 0.5})
        self.ax.grid(True, **grid_kwargs)
        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        if self.ylim is not None:
            self.ax.set_ylim(self.ylim)
        if self.title:
            self.ax.set_title(self.title)
        self.apply_legend()

    # Aplica legendas
    def apply_legend(self) -> None:
        handles, labels = self.ax.get_legend_handles_labels()
        if not handles:
            return
        leg = self.ax.legend(
            loc="upper right",
            frameon=True,
            edgecolor="black",
            fancybox=True,
            fontsize=self.style.get("legend_fontsize", 12),
        )
        frame = leg.get_frame()
        frame.set_facecolor("white")
        frame.set_edgecolor("black")
        frame.set_alpha(1)

    # Aplica cores
    def apply_color(self, idx: int) -> Optional[str]:
        if self.colors is None:
            return None
        if isinstance(self.colors, str):
            return self.colors
        if isinstance(self.colors, (list, tuple)):
            return self.colors[idx % len(self.colors)]
        return None

class TimePlot(BasePlot):
    r"""
    Classe para plotar sinais no domínio do tempo, recebendo um vetor de tempo $t$, e uma lista de sinais $s(t)$.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        t (np.ndarray): Vetor de tempo
        signals (Union[np.ndarray, List[np.ndarray]]): Sinal ou lista de sinais $s(t)$.
        time_unit (str): Unidade de tempo para plotagem ("ms" por padrão, pode ser "s").
        amp_norm (bool): Normalização do sinal para amplitude máxima

    Example:
        - Modulador: ![pageplot](assets/example_modulator_time.svg)
        - Demodulador: ![pageplot](assets/example_demodulator_time.svg)
        - Adição de AWGN ![pageplot](assets/example_noise_time.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 t: np.ndarray,
                 signals: Union[np.ndarray, List[np.ndarray]],
                 time_unit: str = "ms",
                 amp_norm: bool = False,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)

        self.amp_norm = amp_norm

        # Copia os sinais de entrada para evitar modificações no sinal original
        original_signals = signals if isinstance(signals, (list, tuple)) else [signals]
        self.signals = [sig.copy() for sig in original_signals]

        # Unidade de tempo
        self.time_unit = time_unit.lower()
        if self.time_unit == "ms":
            self.t = t * 1e3
        else:
            self.t = t

        # Sinal ou lista de sinais
        if self.labels is None:
            self.labels = [f"Signal {i+1}" for i in range(len(self.signals))]

    def plot(self) -> None:
        # Normalização
        if self.amp_norm:
            max_val = np.max(np.abs(np.concatenate(self.signals)))
            if max_val > 0:
                f = 1 / max_val
                for i, sig in enumerate(self.signals):
                    self.signals[i] *= f

        # Plotagem
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        for i, sig in enumerate(self.signals):
            color = self.apply_color(i)
            if color is not None:
                self.ax.plot(self.t, sig, label=self.labels[i], color=color, **line_kwargs)
            else:
                self.ax.plot(self.t, sig, label=self.labels[i], **line_kwargs)

        # Labels
        xlabel = r"Tempo ($ms$)" if self.time_unit == "ms" else r"Tempo ($s$)"
        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(r"Amplitude")
        self.apply_ax_style()

class FrequencyPlot(BasePlot):
    r"""
    Classe para plotar sinais no domínio da frequência, recebendo uma frequência de amostragem $f_s$ e um sinal $s(t)$ e realizando a transformada de Fourier do sinal, conforme a expressão abaixo. 

    $$
    \begin{equation}
        S(f) = \mathcal{F}\{s(t)\}
    \end{equation}
    $$

    Sendo:
        - $S(f)$: Sinal no domínio da frequência.
        - $s(t)$: Sinal no domínio do tempo.
        - $\mathcal{F}$: Transformada de Fourier.
    
    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        fs (float): Frequência de amostragem
        signal (np.ndarray): Sinal a ser plotado
        fc (float): Frequência central

    Example:
        - Modulador: ![pageplot](assets/example_modulator_freq.svg)
        - Demodulador: ![pageplot](assets/example_demodulator_freq.svg)
        - Adição de AWGN ![pageplot](assets/example_noise_freq.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 fs: float,
                 signal: np.ndarray,
                 fc: float = 0.0,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.fs = fs
        self.fc = fc
        self.signal = signal

    def plot(self) -> None:
        # Transformada de Fourier
        freqs = np.fft.fftshift(np.fft.fftfreq(len(self.signal), d=1 / self.fs))
        fft_signal = np.fft.fftshift(np.fft.fft(self.signal))
        y = mag2db(fft_signal)

        # Escala de frequência
        if self.fc > 1000:
            freqs = freqs / 1000
            self.ax.set_xlabel(r"Frequência ($kHz$)")
        else:
            self.ax.set_xlabel(r"Frequência ($Hz$)")

        # Plotagem
        line_kwargs = {"linewidth": 1, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        color = self.apply_color(0)
        label = self.labels[0] if self.labels else None
        if color is not None:
            self.ax.plot(freqs, y, label=label, color=color, **line_kwargs)
        else:
            self.ax.plot(freqs, y, label=label, **line_kwargs)

        # Labels
        self.ax.set_ylabel(r"Magnitude ($dB$)")
        if self.ylim is None:
            self.ax.set_ylim(-60, 5)

        self.apply_ax_style()

class ConstellationPlot(BasePlot):
    r"""
    Classe para plotar sinais no domínio da constelação, recebendo os sinais $d_I$ e $d_Q$, realizando o plot em fase $I$ e quadratura $Q$, conforme a expressão abaixo.

    $$
    s(t) = d_I(t) + j d_Q(t)
    $$

    Sendo:
        - $s(t)$: Sinal complexo.
        - $d_I(t)$: Sinal em fase.
        - $d_Q(t)$: Sinal em quadratura.


    O plot de constelação pode ser normalizado por um fator de normalização dado por: 

    $$
    \varphi = \frac{\text{A}}{
          \sqrt{
            \displaystyle \frac{1}{N} 
            \sum_{n=0}^{N-1} \Big( I(n)^2 + Q(n)^2 \Big)
          }
        }
    $$

    Sendo:
        - $\text{A}$: Amplitude desejada, definido como `1`. 
        - $\varphi$: Fator de normalização.
        - $N$: Número de amostras.
        - $I(n)$ e $Q(n)$: Sinais em fase e quadratura.
    
    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        dI (np.ndarray): Sinal I
        dQ (np.ndarray): Sinal Q

    Example:
        - Fase e Constelação: ![pageplot](assets/example_modulator_constellation.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 dI: np.ndarray,
                 dQ: np.ndarray,
                 show_ideal_points: bool = True, 
                 rms_norm: bool = False,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.dI = dI
        self.dQ = dQ
        self.amp = 1
        self.show_ideal_points = show_ideal_points
        self.rms_norm = rms_norm

    def plot(self) -> None:
        # Centraliza os dados em torno do zero
        dI_c, dQ_c = self.dI.copy(), self.dQ.copy()
    
        # Se amp_norm for True, normaliza os sinais usando 1/RMS
        if self.rms_norm:
            max_val = np.sqrt(np.mean(dI_c**2 + dQ_c**2))
            if max_val > 0:
                f = self.amp / max_val
                dI_c *= f
                dQ_c *= f
            lim = 1.2 * self.amp
        else:
            lim = 1.2 * np.max(np.abs(np.concatenate([dI_c, dQ_c])))
    
        # Amostras IQ
        scatter_kwargs = {"s": 20, "alpha": 0.6}
        scatter_kwargs.update(self.style.get("scatter", {}))
        color = self.apply_color(0) or "darkgreen"
        self.ax.scatter(dI_c, dQ_c, label="Amostras IQ", color=color, **scatter_kwargs)
    
        # Pontos ideais QPSK
        qpsk_points = np.array([
            [self.amp, self.amp],
            [self.amp, -self.amp],
            [-self.amp, self.amp],
            [-self.amp, -self.amp]
        ])
        if self.show_ideal_points:
            self.ax.scatter(qpsk_points[:, 0], qpsk_points[:, 1],
                            color="blue", s=160, marker="o",
                            label="Pontos Ideais", linewidth=2)
    
        # Linhas auxiliares
        self.ax.axhline(0, color="gray", linestyle="--", alpha=0.5)
        self.ax.axvline(0, color="gray", linestyle="--", alpha=0.5)    
        self.ax.set_xlim(-lim, lim)
        self.ax.set_ylim(-lim, lim)
    
        # Labels
        self.ax.set_xlabel("In Phase $I$")
        self.ax.set_ylabel("Quadrature $Q$")
        self.apply_ax_style()

class BitsPlot(BasePlot):
    r"""
    Classe para plotar bits, recebendo uma lista de bits $b_t$ e realizando o plot em função do tempo $t$.
    
    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        bits_list (List[np.ndarray]): Lista de bits
        sections (Optional[List[Tuple[str, int]]]): Seções do plot
        colors (Optional[List[str]]): Cores do plot
        show_bit_values (bool): Se `True`, exibe os valores dos bits.
        xlabel (Optional[str]): Label do eixo x.
        ylabel (Optional[str]): Label do eixo y.
        label (Optional[str]): Label do plot.
        xlim (Optional[Tuple[float, float]]): Limites do eixo x.

    Example:
        - Datagrama: ![pageplot](assets/example_datagram_time.svg)
        - Codificador Convolucional: ![pageplot](assets/example_conv_time.svg)
        - Embaralhador: ![pageplot](assets/example_scrambler_time.svg)
        - Preâmbulo: ![pageplot](assets/example_preamble.svg)
        - Multiplexador: ![pageplot](assets/example_mux.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 bits_list: List[np.ndarray],
                 sections: Optional[List[Tuple[str, int]]] = None,
                 colors: Optional[List[str]] = None,
                 show_bit_values: bool = True,
                 xlabel: Optional[str] = None,
                 ylabel: Optional[str] = None,
                 label: Optional[str] = None,
                 xlim: Optional[Tuple[float, float]] = None,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.bits_list = bits_list
        self.sections = sections
        self.colors = colors
        self.show_bit_values = show_bit_values
        self.xlim = xlim
        self.bit_value_offset = 0.15
        self.bit_value_size = 13
        self.bit_value_weight = 'bold'
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.label = label

    def plot(self) -> None:
        # Concatena e superamostra os vetores de bit.
        all_bits = np.concatenate(self.bits_list)
        bits_up = np.repeat(all_bits, 2)
        x = np.arange(len(bits_up))

        # Ajustes de eixo
        y_upper = 1.4 if self.show_bit_values else 1.2
        if self.xlim is not None:
            # Ajusta o limite superior do xlim para xlim*2
            self.xlim = (self.xlim[0], self.xlim[1]*2)
            self.ax.set_xlim(self.xlim)
        else:
            self.ax.set_xlim(0, len(bits_up))
        self.ax.set_ylim(-0.2, y_upper)
        self.ax.grid(False)
        self.ax.set_yticks([0, 1])

        # Linhas auxiliares
        self.ax.xaxis.set_major_formatter(FuncFormatter(lambda val, pos: int(val/2)))
        bit_edges = np.arange(0, len(bits_up) + 1, 2)
        for pos in bit_edges:
            self.ax.axvline(x=pos, color='gray', linestyle='--', linewidth=0.5)

        # Para cada vetor de bits, desenha uma seção do plot.
        if self.sections:
            start_bit = 0
            for i, (sec_name, sec_len) in enumerate(self.sections):
                bit_start = start_bit * 2
                bit_end = (start_bit + sec_len) * 2
                color = self.colors[i] if self.colors and i < len(self.colors) else 'black'
                if i > 0:
                    bit_start -= 1

                # Desenha a seção do plot.
                self.ax.step(
                    x[bit_start:bit_end],
                    bits_up[bit_start:bit_end],
                    where='post',
                    color=color,
                    linewidth=2.0,
                    label=sec_name if self.label is None else self.label
                )
                
                # Exibe os valores dos bits acima da linha do plot.
                if self.show_bit_values:
                    xmin, xmax = self.ax.get_xlim()
                    section_bits = all_bits[start_bit:start_bit + sec_len]
                    for j, bit in enumerate(section_bits):
                        xpos = (start_bit + j) * 2 + 1
                        if xpos < xmin or xpos > xmax:
                            continue
                        self.ax.text(
                            xpos,
                            1.0 + self.bit_value_offset,
                            str(int(bit)),
                            ha='center',
                            va='bottom',
                            fontsize=self.bit_value_size,
                            fontweight=self.bit_value_weight,
                            color='black'
                        )
                start_bit += sec_len
        else:
            # Desenha a seção do plot.
            self.ax.step(x, bits_up, where='post',
                         color='black', linewidth=2.0,
                         label=self.label if self.label else None)

            # Exibe os valores dos bits acima da linha do plot.
            if self.show_bit_values:
                xmin, xmax = self.ax.get_xlim()
                for i, bit in enumerate(all_bits):
                    xpos = i * 2 + 1
                    if xpos < xmin or xpos > xmax:
                        continue
                    self.ax.text(
                        xpos,
                        1.0 + self.bit_value_offset,
                        str(int(bit)),
                        ha='center',
                        va='bottom',
                        fontsize=self.bit_value_size,
                        fontweight=self.bit_value_weight
                    )

        # Labels
        if self.xlabel:
            self.ax.set_xlabel(self.xlabel)
        if self.ylabel:
            self.ax.set_ylabel(self.ylabel)
        self.apply_ax_style()

class SymbolsPlot(BasePlot):
    r"""
    Classe para plotar simbolos codificados com codificação de linha, recebendo um vetor de simbolos $s[i]$ e realizando o plot em função do index de simbolo $i$.
    
    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        symbols_list (List[np.ndarray]): Lista de símbolos
        samples_per_symbol (int): Número de amostras por símbolo
        sections (Optional[List[Tuple[str, int]]]): Seções do plot
        colors (Optional[List[str]]): Cores do plot
        show_symbol_values (bool): Se `True`, exibe os valores dos símbolos.
        xlabel (Optional[str]): Label do eixo x.
        ylabel (Optional[str]): Label do eixo y.
        label (Optional[str]): Label do plot.
        xlim (Optional[Tuple[float, float]]): Limites do eixo x.

    Example:
        - Codificação de Canal: ![pageplot](assets/example_encoder_time.svg)
    """

    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 symbols_list: List[np.ndarray],
                 samples_per_symbol: int = 1,
                 sections: Optional[List[Tuple[str, int]]] = None,
                 colors: Optional[List[str]] = None,
                 show_symbol_values: bool = True,
                 xlabel: Optional[str] = None,
                 ylabel: Optional[str] = None,
                 label: Optional[str] = None,
                 xlim: Optional[Tuple[float, float]] = None,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.symbols_list = symbols_list
        self.samples_per_symbol = samples_per_symbol
        self.sections = sections
        self.colors = colors
        self.show_symbol_values = show_symbol_values
        self.xlim = xlim
        self.symbol_value_offset = 0.15
        self.symbol_value_size = 13
        self.symbol_value_weight = 'bold'
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.label = label

    def plot(self) -> None:

        # Concatena e superamostra os vetores de símbolos
        all_symbols = np.concatenate(self.symbols_list)
        symbols_up = np.repeat(all_symbols, self.samples_per_symbol)
        x = np.arange(len(symbols_up))

        # Ajustes de eixo
        y_upper = 1.8 if self.show_symbol_values else 1.5
        if self.xlim is not None:
            self.xlim = (self.xlim[0] * self.samples_per_symbol,
                         self.xlim[1] * self.samples_per_symbol)
            self.ax.set_xlim(self.xlim)
        else:
            self.ax.set_xlim(0, len(symbols_up))
        self.ax.set_ylim(-1.5, y_upper)
        self.ax.set_yticks([-1, 0, 1])
        self.ax.grid(False)

        # Linhas verticais marcando início de cada símbolo
        self.ax.xaxis.set_major_formatter(FuncFormatter(lambda val, pos: int(val / self.samples_per_symbol)))
        symbol_edges = np.arange(0, len(symbols_up) + 1, self.samples_per_symbol)
        for pos in symbol_edges:
            self.ax.axvline(x=pos, color='gray', linestyle='--', linewidth=0.5)

        # Para cada vetor de símbolos, desenha uma seção do plot
        if self.sections:
            start_symbol = 0
            for i, (sec_name, sec_len) in enumerate(self.sections):
                sym_start = start_symbol * self.samples_per_symbol
                sym_end = (start_symbol + sec_len) * self.samples_per_symbol
                color = self.colors[i] if self.colors and i < len(self.colors) else 'black'
                if i > 0:
                    sym_start -= 1

                # Desenha a seção do plot.
                self.ax.step(
                    x[sym_start:sym_end],
                    symbols_up[sym_start:sym_end],
                    where='post',
                    color=color,
                    linewidth=2.0,
                    label=sec_name if self.label is None else self.label
                )

                # Exibe os valores dos símbolos acima da linha do plot.
                if self.show_symbol_values:
                    xmin, xmax = self.ax.get_xlim()
                    section_symbols = all_symbols[start_symbol:start_symbol + sec_len]
                    for j, sym in enumerate(section_symbols):
                        xpos = (start_symbol + j) * self.samples_per_symbol + 0.5 * self.samples_per_symbol
                        if xpos < xmin or xpos > xmax:
                            continue
                        self.ax.text(
                            xpos,
                            1.0 + self.symbol_value_offset,
                            str(int(sym)),
                            ha='center',
                            va='bottom',
                            fontsize=self.symbol_value_size,
                            fontweight=self.symbol_value_weight,
                            color='black'
                        )
                start_symbol += sec_len
        else:

            # Desenha a seção do plot.
            color = self.colors[0] if self.colors else 'black'
            self.ax.step(
                x, symbols_up, where='post',
                color=color, linewidth=2.0,
                label=self.label if self.label else None
            )
            # Exibe os valores dos símbolos acima da linha do plot.
            if self.show_symbol_values:
                xmin, xmax = self.ax.get_xlim()
                for i, sym in enumerate(all_symbols):
                    xpos = i * self.samples_per_symbol + 0.5 * self.samples_per_symbol
                    if xpos < xmin or xpos > xmax:
                        continue
                    self.ax.text(
                        xpos,
                        1.0 + self.symbol_value_offset,
                        str(int(sym)),
                        ha='center',
                        va='bottom',
                        fontsize=self.symbol_value_size,
                        fontweight=self.symbol_value_weight
                    )

        # Labels
        if self.xlabel:
            self.ax.set_xlabel(self.xlabel)
        if self.ylabel:
            self.ax.set_ylabel(self.ylabel)
        self.apply_ax_style()

class ImpulseResponsePlot(BasePlot):
    r"""
    Classe para plotar a resposta ao impulso de um filtro, recebendo um vetor de tempo $t_{imp}$ e realizando o plot em função do tempo $t$.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot no GridSpec
        t_imp (np.ndarray): Vetor de tempo da resposta ao impulso
        impulse_response (np.ndarray): Amostras da resposta ao impulso
        t_unit (str, optional): Unidade de tempo no eixo X ("ms" ou "s"). Default é "ms"
        label (Optional[Union[str, List[str]]]): Label do plot
        xlabel (Optional[str]): Label do eixo x
        ylabel (Optional[str]): Label do eixo y
        xlim (Optional[Tuple[float, float]]): Limites do eixo x
        amp_norm (Optional[bool]): Normaliza a resposta ao impulso para ter amplitude unitária. 

    Example:
        - Resposta ao Impulso RRC: ![pageplot](assets/example_formatter_impulse.svg)
        - Resposta ao Impulso Filtro Passa baixa: ![pageplot](assets/example_lpf_impulse.svg)
        - Resposta ao Impulso RRC Invertido: ![pageplot](assets/example_mf_impulse.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 t_imp: np.ndarray,
                 impulse_response: Union[np.ndarray, List[np.ndarray]],
                 t_unit: str = "ms",
                 label: Optional[Union[str, List[str]]] = None,
                 xlabel: Optional[str] = None,
                 ylabel: Optional[str] = None,
                 xlim: Optional[Tuple[float, float]] = None,
                 amp_norm: Optional[bool] = False,
                 **kwargs) -> None:

        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.t_imp = t_imp
        self.label = label
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xlim = xlim
        self.amp_norm = amp_norm
        self.t_unit = t_unit

        # Cria lista de respostas ao impulso
        if isinstance(impulse_response, np.ndarray):
            self.impulse_response = [impulse_response]
        else:
            self.impulse_response = impulse_response


    def plot(self) -> None:
        # Unidade de tempo
        if self.t_unit == "ms":
            t_plot = self.t_imp * 1000
            default_xlabel = r"Tempo ($ms$)"
        else:
            t_plot = self.t_imp
            default_xlabel = r"Tempo ($s$)"

        # Label
        if isinstance(self.label, str) or self.label is None:
            labels = [self.label] * len(self.impulse_response)
        else:
            labels = self.label
        self.ax.set_xlabel(self.xlabel if self.xlabel is not None else default_xlabel)
        self.ax.set_ylabel(self.ylabel if self.ylabel is not None else "Amplitude")

        # Plotagem
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        for i, resp in enumerate(self.impulse_response):
            color = self.apply_color(i) or None
            lbl = labels[i] if labels and i < len(labels) else None
            if self.amp_norm:
                resp = resp / np.max(resp)
            self.ax.plot(t_plot, resp, color=color, label=lbl, **line_kwargs)

        # Limites
        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        self.apply_ax_style()

class SampledSignalPlot(BasePlot):
    r"""
    Classe para plotar um sinal $s(t)$ amostrado em $t_s$.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int ou tuple): Posição no GridSpec
        t_signal (np.ndarray): Vetor de tempo do sinal filtrado
        signal (np.ndarray): Sinal filtrado
        t_samples (np.ndarray): Instantes de amostragem
        samples (np.ndarray): Amostras correspondentes
        time_unit (str): Unidade de tempo. 
        label_signal (str): Label do sinal filtrado.
        label_samples (str): Label das amostras.
        xlabel (str): Label do eixo x.
        ylabel (str): Label do eixo y.
        title (str): Título do plot.
        xlim (tuple): Limites do eixo x.

    Example:
        ![pageplot](assets/example_sampler_time.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 t_signal: np.ndarray,
                 signal: np.ndarray,
                 t_samples: np.ndarray,
                 samples: np.ndarray,
                 time_unit: str = "ms",
                 label_signal: Optional[str] = None,
                 label_samples: Optional[str] = None,
                 xlabel: Optional[str] = None,
                 ylabel: Optional[str] = "Amplitude",
                 title: Optional[str] = None,
                 xlim: Optional[Tuple[float, float]] = None,
                 **kwargs) -> None:

        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)

        self.time_unit = time_unit.lower()
        self.label_signal = label_signal
        self.label_samples = label_samples
        if xlabel is None:
            xlabel = r"Tempo ($ms$)" if self.time_unit == "ms" else r"Tempo ($s$)"
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title = title
        self.xlim = xlim
        self.signal = signal
        self.samples = samples

        # Ajusta unidade de tempo
        if self.time_unit == "ms":
            self.t_signal = t_signal * 1e3
            self.t_samples = t_samples * 1e3
        else:
            self.t_signal = t_signal
            self.t_samples = t_samples

    def plot(self) -> None:
        # Plotagem
        signal_color = self.colors if isinstance(self.colors, str) else "blue"
        self.ax.plot(self.t_signal, self.signal,color=signal_color, label=self.label_signal, linewidth=2)
        self.ax.stem(self.t_samples, self.samples,linefmt="k-", markerfmt="ko", basefmt=" ",label=self.label_samples)

        # Ajuste dos eixos
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.ax.set_xlim(self.xlim)
        self.ax.set_title(self.title)
        self.apply_ax_style()
        
        # Legenda
        if self.label_signal or self.label_samples:
            leg = self.ax.legend(loc='upper right', frameon=True, fontsize=12)
            leg.get_frame().set_facecolor("white")
            leg.get_frame().set_edgecolor("black")
            leg.get_frame().set_alpha(1.0)

class PhasePlot(BasePlot):
    r"""
    Classe para plotar a fase dos sinais $d_I(t)$ e $d_Q(t)$ no domínio do tempo.

    $$
        s(t) = \arctan\left(\frac{d_Q(t)}{d_I(t)}\right)
    $$

    Sendo: 
        - $s(t)$: Vetor de fases por intervalo de tempo.
        - $d_I(t)$: Componente sinal $d_I(t)$, em fase. 
        - $d_Q(t)$: Componente sinal $d_Q(t)$, em quadratura.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        t (np.ndarray): Vetor de tempo
        signals (Union[np.ndarray, List[np.ndarray]]): Sinais IQ (I e Q)
        time_unit (str): Unidade de tempo para plotagem ("ms" por padrão, pode ser "s").
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos: int,
                 t: np.ndarray,
                 signals: Union[np.ndarray, List[np.ndarray]],
                 time_unit: str = "ms",
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)

        # Unidade de tempo
        self.time_unit = time_unit.lower()
        self.t = t
        if self.time_unit == "ms":
            self.t *= 1e3

        if self.labels is None:
            self.labels = ["Fase IQ"]

        # Garantir que os sinais estão em uma tupla
        if isinstance(signals, (list, tuple)):
            assert len(signals) == 2, "Os sinais devem ser passados como tupla com dois componentes (I, Q)."
            self.I = signals[0]
            self.Q = signals[1]
        else:
            raise ValueError("Os sinais devem ser passados como tupla com dois componentes (I, Q).")

    def plot(self) -> None:
        # Calcula a fase usando atan2
        fase = np.angle(self.I + 1j * self.Q)

        # Plot da fase ao longo do tempo
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        color = self.apply_color(0)
        self.ax.plot(self.t, fase, label=self.labels[0], color=color, **line_kwargs)

        # Limite de fase entre pi e -pi
        self.ax.set_ylim([-np.pi*1.1, np.pi*1.1])
        ticks = [0, np.pi/4, 3*np.pi/4, -np.pi/4, -3*np.pi/4, -np.pi, np.pi]
        labels = [r"$0$", r"$\frac{\pi}{4}$", r"$\frac{3\pi}{4}$", r"$-\frac{\pi}{4}$", r"$-\frac{3\pi}{4}$", r"$-\pi$", r"$\pi$"]
        self.ax.set_yticks(ticks)
        self.ax.set_yticklabels(labels)

        # Ajuste dos eixos
        self.ax.set_xlabel(r"Tempo ($ms$)" if self.time_unit == "ms" else r"Tempo ($s$)")
        self.ax.set_ylabel(r"Fase ($rad$)")
        self.ax.legend()
        self.apply_ax_style()

class GaussianNoisePlot(BasePlot):
    r"""
    Classe para plotar a densidade de probabilidade $p(x)$ de uma dada variância $\sigma^2$, seguindo a expressão abaixo. 

    $$
    p(x) = \frac{1}{\sqrt{2\pi\sigma^2}} \exp\left(-\frac{x^2}{2\sigma^2}\right)
    $$

    Sendo: 
        - $p(x)$: Densidade de probabilidade do ruído.
        - $\sigma^2$: Variância do ruído.
        - $x$: Amplitude do ruído.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot no GridSpec
        variance (float): Variância do ruído
        num_points (int): Número de pontos para a curva da gaussiana
        legend (str): Legenda do plot
        xlabel (str): Label do eixo x
        ylabel (str): Label do eixo y
        xlim (Optional[Tuple[float, float]]): Limites do eixo x
        span (int): Span do plot

    Example:
        ![pageplot](assets/example_noise_gaussian_ebn0.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 variance: float,
                 num_points: int = 5000,
                 legend: str = "Ruído AWGN",
                 xlabel: str = "Amplitude",
                 ylabel: str = "Densidade de Probabilidade",
                 xlim: Optional[Tuple[float, float]] = None,
                 span: int = 100,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.variance = variance
        self.num_points = num_points
        self.legend = legend
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xlim = xlim
        self.span = span

    def plot(self) -> None:
        # Calculo da pdf
        sigma = np.sqrt(self.variance)
        x = np.linspace(-self.span*sigma, self.span*sigma, self.num_points)
        pdf = (1 / (np.sqrt(2*np.pi) * sigma)) * np.exp(-x**2 / (2*self.variance))

        # Plotagem
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        color = self.apply_color(0) or "darkgreen"
        self.ax.plot(x, pdf, label=self.legend, color=color, **line_kwargs)
       
        # Ajuste dos eixos
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        self.apply_ax_style()

class PoleZeroPlot(BasePlot):
    r"""
    Classe para plotar o diagrama de polos e zeros de uma função de transferência discreta no plano-z.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição no GridSpec
        b (np.ndarray): Coeficientes do numerador da função de transferência
        a (np.ndarray): Coeficientes do denominador da função de transferência

    Example:
        ![pageplot](assets/example_lpf_pz.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 b: np.ndarray,
                 a: np.ndarray,
                 **kwargs) -> None:

        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.b = b
        self.a = a

    def plot(self) -> None:
        # Calcula zeros e polos
        zeros = np.roots(self.b)
        poles = np.roots(self.a)

        # Plot da circuferência, polos e zeros
        theta = np.linspace(0, 2*np.pi, 512)
        self.ax.plot(np.cos(theta), np.sin(theta), 'k--', alpha=0.6)
        self.ax.scatter(np.real(zeros), np.imag(zeros),marker='o', facecolors='none', edgecolors='blue', s=120, label='Zeros')
        self.ax.scatter(np.real(poles), np.imag(poles),marker='x', color='red',s=120, label='Polos')

        # Eixos
        self.ax.axhline(0, color='black', linewidth=0.8)
        self.ax.axvline(0, color='black', linewidth=0.8)
        self.ax.set_aspect('equal', adjustable='box')
        self.ax.set_xlim([-1.2, 1.2])
        self.ax.set_ylim([-1.2, 1.2])

        # Labels
        self.ax.set_xlabel("Parte Real")
        self.ax.set_ylabel("Parte Imaginária")
        self.apply_ax_style()

class FrequencyResponsePlot(BasePlot):
    r"""
    Classe para plotar a resposta em frequência de um filtro a partir de seus coeficientes (b, a). 
    Calcula a transformada de Fourier discreta da resposta ao impulso usando `scipy.signal.freqz`.

    $$
        H(f) = \sum_{n=0}^{N} b_n e^{-j 2 \pi f n} \Big/ \sum_{m=0}^{M} a_m e^{-j 2 \pi f m}
    $$

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot no GridSpec
        b (np.ndarray): Coeficientes do numerador do filtro
        a (np.ndarray): Coeficientes do denominador do filtro
        fs (float): Frequência de amostragem
        f_cut (Optional[float]): Frequência de corte do filtro (Hz)
        xlim (Optional[Tuple[float, float]]): Limites do eixo X (Hz)
        worN (int): Número de pontos para a transformada de Fourier
        show_phase (bool): Se `True`, plota a fase da resposta em frequência
        xlabel (str): Label do eixo X
        ylabel (str): Label do eixo Y

    Example:
        ![pageplot](assets/example_lpf_freq_response.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 b: np.ndarray,
                 a: np.ndarray,
                 fs: float,
                 f_cut: float = None,
                 xlim: tuple = None,
                 worN: int = 1024,
                 show_phase: bool = False,
                 xlabel: str = r"Frequência ($Hz$)",
                 ylabel: str = r"Magnitude ($dB$)",
                 **kwargs) -> None:

        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.b = b
        self.a = a
        self.fs = fs
        self.f_cut = f_cut
        self.xlim = xlim
        self.worN = worN
        self.show_phase = show_phase
        self.xlabel = xlabel
        self.ylabel = ylabel    

    def plot(self) -> None:
        # calcula resposta em frequência
        w, h = freqz(self.b, self.a, worN=self.worN, fs=self.fs)
        magnitude = mag2db(h)

        # Plotagem
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        color = self.apply_color(0) or "darkorange"
        label = self.labels[0] if self.labels else "$H(f)$"
        self.ax.plot(w, magnitude, color=color, label=label, **line_kwargs)

        # Plotagem da fase
        if self.show_phase:
            ax2 = self.ax.twinx()
            phase = np.unwrap(np.angle(h))
            ax2.plot(w, phase, color="darkorange", linestyle="--", linewidth=1.5, label="Fase")
            ax2.set_ylabel("Fase (rad)")

        # adiciona a barra vertical na frequência de corte
        if self.f_cut is not None:
            self.ax.axvline(self.f_cut, color="red", linestyle="--", linewidth=2, label=f"$f_c$ = {self.f_cut} Hz")

        # Eixos
        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        self.ax.set_ylim(-60, 5)

        # Labels
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.apply_ax_style()

class DetectionFrequencyPlot(BasePlot):
    r"""
    Classe para plotar o espectro de uma sinal recebido, com threshold e frequências detectadas. Recebendo uma frequência de amostragem $f_s$ e um sinal $s(t)$ e realizando a transformada de Fourier do sinal, conforme a expressão abaixo. 

    $$
    \begin{equation}
        S(f) = \mathcal{F}\{s(t)\}
    \end{equation}
    $$

    Sendo:
        - $S(f)$: Sinal no domínio da frequência.
        - $s(t)$: Sinal no domínio do tempo.
        - $\mathcal{F}$: Transformada de Fourier.
    
    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição do plot
        fs (float): Frequência de amostragem
        signal (np.ndarray): Sinal a ser plotado
        fc (float): Frequência central

    Example: 
        ![pageplot](assets/example_detector_freq.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 fs: float,
                 signal: np.ndarray,
                 threshold: float,
                 fc: float = 0.0,
                 title: str = "",
                 labels: Optional[List[str]] = None,
                 xlim: Optional[Tuple[float, float]] = None,
                 ylim: Optional[Tuple[float, float]] = None,
                 colors: Optional[Union[str, List[str]]] = None,
                 style: Optional[Dict[str, Any]] = None,
                 freqs_detected: Optional[Union[List[float], np.ndarray]] = None
                 ) -> None:

        ax = fig.add_subplot(grid[pos])
        super().__init__(ax,
                         title=title,
                         labels=labels,
                         xlim=xlim,
                         ylim=ylim,
                         colors=colors,
                         style=style)

        self.fs = fs
        self.fc = fc
        self.signal = np.asarray(signal)
        self.threshold = threshold
        self.freqs_detected = freqs_detected
        self.U = 1.0
        self.style = self.style or {}

    def plot(self) -> None:
        P_db = self.signal
        if P_db.ndim != 1:
            raise ValueError("DetectionFrequencyPlot espera um vetor de power_matrix.")

        n_bins = len(P_db)
        n_fft = 2 * (n_bins - 1)
        freqs = np.fft.rfftfreq(n_fft, d=1.0 / self.fs)

        # Plotagem em KHz
        freqs_plot = freqs / 1000.0
        line_kwargs = {"linewidth": 1.5, "alpha": 0.9}
        line_kwargs.update(self.style.get("line", {}))
        color = self.apply_color(0) or "blue"
        label = self.labels[0] if self.labels else "Espectro (dB)"
        self.ax.plot(freqs_plot, P_db, label=label, color=color, **line_kwargs)

        # Threshold
        thr_line = self.threshold
        thr_label = f"Threshold = {thr_line:.2f} dB"
        self.ax.axhline(thr_line, color="blue", linestyle="--", linewidth=2, label=thr_label)

        # Plota as linhas verticais
        detected_bins = np.where(np.asarray(self.freqs_detected) > 0)[0]
        for idx, k in enumerate(detected_bins, start=1):
            f_plot = freqs[k] / 1000.0
            Pk = P_db[k]
            self.ax.plot(f_plot, Pk, 'o', color='k', markersize=6, label=f"$f_{{{idx}}} = {f_plot:.2f}$ kHz")
            self.ax.axvline(f_plot, color='k', linestyle=':', linewidth=2)

        # Limites dos eixos
        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        if self.ylim is not None:
            self.ax.set_ylim(self.ylim)

        # Labels
        handles, labels = self.ax.get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        self.ax.legend(by_label.values(), by_label.keys())
        self.ax.set_xlabel(r"Frequência ($kHz$)")
        self.ax.set_ylabel(r"Magnitude ($dB$)")
        self.apply_ax_style()

class BersnrPlot(BasePlot):
    r"""
    Classe para plotar curvas de BER em função de Eb/N0.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int): Posição no GridSpec
        EbN0 (np.ndarray): Vetor de valores Eb/N0 (dB)
        ber_curves (List[np.ndarray]): Lista de curvas BER correspondentes
        labels (List[str]): Rótulos de cada curva
        linestyles (List[str], opcional): Lista com estilos de linha. 
        markers (List[str], opcional): Lista com formatos de marcadores. 
        xlabel (str, opcional): Rótulo do eixo x
        ylabel (str, opcional): Rótulo do eixo y
        logy (bool, opcional): Se deve usar escala logarítmica no eixo y

    Example: 
        - ![pageplot](assets/ber_vs_ebn0.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos: int,
                 EbN0: np.ndarray,
                 ber_curves: List[np.ndarray],
                 linestyles: List[str] = None,
                 markers: List[str] = None,
                 xlabel: str = r"$E_b/N_0$ (dB)",
                 ylabel: str = r"Taxa de Erro de Bit (BER)",
                 logy: bool = True,
                 **kwargs) -> None:

        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)

        self.xlabel = xlabel
        self.ylabel = ylabel
        self.logy = logy
        self.EbN0 = EbN0

        # Cria as curvas BER
        self.ber_curves = ber_curves if isinstance(ber_curves, (list, tuple)) else [ber_curves]

        # Cria os rótulos
        if self.labels is None:
            self.labels = [f"Curva {i+1}" for i in range(len(self.ber_curves))]
        self.linestyles = linestyles if linestyles is not None else ["-"] * len(self.ber_curves)
        self.markers = markers if markers is not None else ["o"] * len(self.ber_curves)

    def plot(self) -> None:
        # Plotagem
        for i, curve in enumerate(self.ber_curves):
            color = self.apply_color(i)
            label = self.labels[i]
            linestyle = self.linestyles[i % len(self.linestyles)]
            marker = self.markers[i % len(self.markers)]

            plot_kwargs = {"linewidth": 2, "alpha": 1.0,
                           "linestyle": linestyle,
                           "marker": marker}

            self.ax.plot(self.EbN0, curve, label=label, color=color, **plot_kwargs)

        # Usa escala logarítmica por padrão
        if self.logy:
            self.ax.set_yscale("log")
            self.ax.grid(True, which="both", axis="y", linestyle="--", color="gray", alpha=0.6)

        # Labels
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.apply_ax_style()

class SincronizationPlot(BasePlot):
    r"""
    Classe para plotar um sinal no domínio do tempo com marcações de sincronismo.

    Args:
        fig (plt.Figure): Figura do plot
        grid (gridspec.GridSpec): GridSpec do plot
        pos (int ou tuple): Posição no GridSpec
        t (np.ndarray): Vetor de tempo
        signal (np.ndarray): Sinal no tempo
        sync_start (float): Instante de início da palavra de sincronismo
        sync_end (float): Instante de fim da palavra de sincronismo
        max_corr (float): Instante do pico de correlação
        time_unit (str): Unidade de tempo para plotagem ("ms" por padrão, pode ser "s").

    Example: 
        ![pageplot](assets/example_synchronizer_sync.svg)
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 t: np.ndarray,
                 signal: np.ndarray,
                 sync_start: float,
                 sync_end: float,
                 max_corr: float,
                 time_unit: str = "ms",
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)

        # Ajuste de unidade de tempo
        self.time_unit = time_unit.lower()
        if self.time_unit == "ms":
            self.t = t * 1e3
            self.sync_start = sync_start * 1e3
            self.sync_end = sync_end * 1e3
            self.max_corr = max_corr * 1e3
        else:
            self.t = t
            self.sync_start = sync_start
            self.sync_end = sync_end
            self.max_corr = max_corr

        self.signals = [signal]
        if self.labels is None:
            self.labels = ["Signal"]

    def plot(self) -> None:
        # Plotagem
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        for i, sig in enumerate(self.signals):
            color = self.apply_color(i)
            if color is not None:
                self.ax.plot(self.t, sig, label=self.labels[i], color=color, **line_kwargs)
            else:
                self.ax.plot(self.t, sig, label=self.labels[i], **line_kwargs)

        # Periodo de recepção do preambulo
        self.ax.axvspan(self.sync_start, self.sync_end,
                        color="gray", alpha=0.2, label=r"$\Delta \tau$")

        # Linhas verticais de sincronismo
        self.ax.axvline(self.max_corr, color="darkorange", linestyle="--", linewidth=2, label=r"$\tau$")
        self.ax.axvline(self.sync_start, color="red", linestyle="--", linewidth=2, label=r"$\tau +/- (\Delta \tau)/2$")
        self.ax.axvline(self.sync_end, color="red", linestyle="--", linewidth=2)
    
        # Labels
        xlabel = r"Tempo ($ms$)" if self.time_unit == "ms" else r"Tempo ($s$)"
        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(r"Amplitude")
        self.apply_ax_style()

class CorrelationPlot(BasePlot):
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 corr_vec: np.ndarray,
                 fs: float,
                 xlim_ms: Tuple[float, float],
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        
        # Encontrar o índice de maior correlação
        self.corr_vec = corr_vec
        self.sample_indices = np.arange(len(corr_vec))
        self.fs = fs
        self.max_corr_index = np.argmax(corr_vec)
        self.max_corr_value = corr_vec[self.max_corr_index]

        # Limites em índices de amostra
        self.index_low = int(xlim_ms[0] * 1e-3 * fs)
        self.index_high = int(xlim_ms[1] * 1e-3 * fs)

        # Definir o título e a legenda
        if self.labels is None:
            self.labels = [r"$c[k]$"]

    def plot(self) -> None:
        # Plotagem
        color = self.apply_color(0)
        line_kwargs = {"linewidth": 2, "alpha": 1.0}
        line_kwargs.update(self.style.get("line", {}))
        self.ax.plot(self.sample_indices, self.corr_vec, label=self.labels[0], color=color, **line_kwargs)

        # Máxima correlação
        self.ax.axvline(self.max_corr_index, color='red', linestyle='--', label=f"$k_{{max}}$ = {self.max_corr_index}")
        self.ax.scatter(self.max_corr_index, self.max_corr_value, color='red', zorder=5)

        # limites
        self.ax.set_xlim(self.index_low, self.index_high)

        # Labels
        self.ax.set_xlabel(r"Índice de Amostra $k$")
        self.ax.set_ylabel(r"Fator de Correlação Normalizado $c[k]$")
        self.ax.legend()
        self.apply_ax_style()

class PowerMatrixPlot(BasePlot):
    r"""
    Plota a matriz de potência em dB como heatmap quadriculado.
    Eixo x = segmentos, eixo y = frequência em Hz.
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 power_matrix: np.ndarray,
                 fs: float,
                 N: int,
                 xlim: Tuple[float, float] = (0, 10),
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.power_matrix = power_matrix
        self.fs = fs
        self.N = N
        self.xlim = xlim

    def plot(self) -> None:
        n_segments, n_freqs = self.power_matrix.shape

        # Frequências em kHz
        freqs = np.fft.rfftfreq(self.N, d=1/self.fs)
        freqs_khz = freqs / 1000.0
        x = np.linspace(freqs_khz[0], freqs_khz[-1], n_freqs + 1)

        # Segmentos no eixo Y
        y = np.arange(n_segments + 1)

        # Plotagem
        im = self.ax.pcolormesh(
            x, y, self.power_matrix,
            cmap="inferno", shading="auto"
        )
        self.ax.invert_yaxis()

        # Barra de cores    
        cbar = self.ax.figure.colorbar(im, ax=self.ax)
        cbar.set_label("Magnitude ($dB$)")

        # Limites de frequência no eixo X
        self.ax.set_xlim(self.xlim[0], self.xlim[1])

        # Labels
        self.ax.set_xlabel("Frequência ($kHz$)")
        self.ax.set_ylabel("Índice de Segmento (tempo)")
        self.ax.grid(False)
        self.apply_ax_style()



class PowerMatrix3DPlot(BasePlot):
    r"""
    Plota a matriz de potência em dB em 3D, limitando a frequência
    ao range definido em kHz e adicionando plano de threshold.
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 power_matrix: np.ndarray,
                 fs: float,
                 N: int,
                 freq_window: tuple[float, float] = (0, 10),
                 threshold: float = None,
                 smooth: bool = True,
                 sigma: float = 1.0,
                 elev: float = 5.0,
                 azim: float = -60.0,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos], projection="3d")
        super().__init__(ax, **kwargs)
        self.power_matrix = power_matrix
        self.fs = fs
        self.N = N
        self.freq_window = freq_window
        self.threshold = threshold
        self.smooth = smooth
        self.sigma = sigma
        self.elev = elev
        self.azim = azim

    def plot(self) -> None:
        n_segments, n_freqs = self.power_matrix.shape
        freqs = np.fft.rfftfreq(self.N, d=1/self.fs)

        # aplica janela de frequências
        if self.freq_window is not None:
            fmin, fmax = self.freq_window
            mask = (freqs >= fmin) & (freqs <= fmax)
            freqs = freqs[mask]
            Z = self.power_matrix[:, mask]
        else:
            Z = self.power_matrix

        # aplica suavização (apenas para ficar mais legivel)
        if self.smooth:
            Z = gaussian_filter(Z, sigma=self.sigma)

        X = np.arange(Z.shape[0])
        Y = freqs / 1000.0
        X, Y = np.meshgrid(X, Y, indexing="ij")

        # superfície da matriz de potência
        surf = self.ax.plot_surface(
            X, Y, Z,
            cmap="inferno",
            linewidth=0,
            antialiased=True,
            alpha=0.95
        )

        # plano do threshold
        if self.threshold is not None:
            Z_thr = np.full_like(Z, self.threshold)
            self.ax.plot_surface(
                X, Y, Z_thr,
                color="blue", alpha=0.5, rstride=1, cstride=1, linewidth=0
            )

        self.ax.set_xlabel("Segmento $n$", labelpad=15)
        self.ax.set_ylabel("Frequência ($kHz$)", labelpad=15)
        self.ax.set_zlabel("Magnitude ($dB$)", labelpad=15)

        # diminui da legenda do eixo do segmento para ficar mais legivel
        self.ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(3))

        if self.freq_window is not None:
            self.ax.set_ylim(self.freq_window[0]/1000, self.freq_window[1]/1000)

        # aplica ângulo da câmera
        self.ax.view_init(elev=self.elev, azim=self.azim)

class MatrixSquarePlot(BasePlot):
    r"""
    Plota matrizes categóricas (detecção/decisão) em formato quadriculado.
    Eixo x = frequência (kHz), eixo y = segmentos (tempo).
    """
    def __init__(self,
                 fig: plt.Figure,
                 grid: gridspec.GridSpec,
                 pos,
                 matrix: np.ndarray,
                 fs: float,
                 N: int,
                 xlim: Tuple[float, float] = (0, 10),
                 legend_list: List[str] = None,
                 **kwargs) -> None:
        ax = fig.add_subplot(grid[pos])
        super().__init__(ax, **kwargs)
        self.matrix = matrix
        self.fs = fs
        self.N = N
        self.xlim = xlim
        self.legend_list = legend_list or ["Detectada", "Confirmada", "Span", "Demodulação"]

        self.cmap = mpl.colors.ListedColormap([
            (1, 1, 1, 0),
            "blue",
            "red",
            "lightblue",
            "orange"
        ])
        self.bounds = [0, 1, 2, 3, 4, 5]
        self.norm = mpl.colors.BoundaryNorm(self.bounds, self.cmap.N)

    def plot(self) -> None:
        n_segments, n_freqs = self.matrix.shape

        # Eixo X = frequências (kHz) -> deve ter comprimento n_freqs + 1
        freqs = np.fft.rfftfreq(self.N, d=1/self.fs)
        freqs_khz = freqs / 1000.0
        x = np.linspace(freqs_khz[0], freqs_khz[-1], n_freqs + 1)

        # Eixo Y = segmentos -> deve ter comprimento n_segments + 1
        y = np.arange(n_segments + 1)

        # Agora matrix tem shape (n_segments, n_freqs) -> compatível com (len(y)-1, len(x)-1)
        im = self.ax.pcolormesh(
            x, y, self.matrix,
            cmap=self.cmap,
            norm=self.norm,
            shading="auto"
        )

        # Legenda (categorias)
        legend_map = {
            "Detectada": "blue",
            "Confirmada": "red",
            "Span": "lightblue",
            "Demodulação": "orange",
        }

        legend_elements = [
            Line2D([0], [0],
                   marker='s',
                   color='w',
                   markerfacecolor=color,
                   markersize=12,
                   label=label)
            for label, color in legend_map.items()
            if label in self.legend_list
        ]

        if legend_elements:
            leg = self.ax.legend(handles=legend_elements, loc="upper right")
            frame = leg.get_frame()
            frame.set_edgecolor("black")
            frame.set_alpha(1)

        # Labels e limites
        self.ax.set_xlabel("Frequência ($kHz$)")
        self.ax.set_ylabel("Índice de Segmento (tempo)")
        self.ax.grid(False)

        # Limita frequência no eixo X (já em kHz)
        self.ax.set_xlim(self.xlim[0], self.xlim[1])

        # Inverte o eixo Y (segmento 0 no topo)
        self.ax.invert_yaxis()

        self.apply_ax_style()






# """
# Implementa uma palavra de sincronismo compatível com o padrão PPT-A3.

# Referência:
#     AS3-SP-516-274-CNES (seção 3.1.4.6)

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
from .plotter import save_figure, create_figure, BitsPlot

class Preamble:

    def __init__(self, preamble_hex="2BEEEEBF"):
        r"""
        Gera uma palavra de sincronismo, $S = 2BEEEEBF_{16}$ no padrão ARGOS-3. A palavra de sincronismo composta por 30 bits, $S = [S_0, S_1, S_2, \dots, S_{29}]$ que são intercalados para formar os vetores $S_I$ e $S_Q$ de cada canal, conforme apresentado abaixo.

        $$
        \begin{align}
        S_I &= [S_0,\, S_2,\, S_4,\, \dots,\, S_{28}] && \mapsto \quad S_I = [1111,\, 1111,\, 1111,\, 111] \\
        S_Q &= [S_1,\, S_3,\, S_5,\, \dots,\, S_{29}] && \mapsto \quad S_Q = [0011,\, 0101,\, 0100,\, 111]
        \end{align}
        $$

        Sendo:
            - $S$: Palavra de sincronismo original.
            - $S_I$ e $S_Q$: Vetores de saida correspondentes aos canais I e Q, respectivamente.

        Args:
            preamble_hex (str, opcional): Hexadecimal da palavra de sincronismo.
        
        Raises:
            ValueError: Se a palavra de sincronismo $S$ tiver comprimento diferente de 8 caracteres. 
            ValueError: Se o hexadecimal não for válido ou não puder ser convertido.

        Example: 
            ![pageplot](assets/example_preamble.svg)

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.6)
        </div>
        """

        if not isinstance(preamble_hex, str) or len(preamble_hex) != 8:
            raise ValueError("O hexadecimal da palavra de sincronismo deve ser uma string de 8 caracteres.")

        self.preamble_hex = preamble_hex
        self.preamble_bits = self.hex_to_bits(self.preamble_hex)

        if len(self.preamble_bits) != 30:
            raise ValueError("A palavra de sincronismo deve conter 30 bits.")

        self.preamble_sI, self.preamble_sQ = self.generate_preamble()

    def hex_to_bits(self, hex_string):
        return format(int(hex_string, 16), '032b')[2:] 
    
    def generate_preamble(self):
        r"""
        Gera os vetores $S_I$ e $S_Q$ da palavra de sincronismo, com base no vetor $S$ passado no construtor.

        Returns:
            tuple (np.ndarray, np.ndarray): Vetores $S_I$ e $S_Q$.
        """
        Si = np.array([int(bit) for bit in self.preamble_bits[::2]])
        Sq = np.array([int(bit) for bit in self.preamble_bits[1::2]])
        return Si, Sq

if __name__ == "__main__":

    preamble = Preamble(preamble_hex="2BEEEEBF")
    Si = preamble.preamble_sI
    Sq = preamble.preamble_sQ

    print("Si: ", ''.join(str(int(b)) for b in Si))
    print("Sq: ", ''.join(str(int(b)) for b in Sq))

    fig_preamble, grid_preamble = create_figure(2, 1, figsize=(16, 9))

    BitsPlot(
        fig_preamble, grid_preamble, (0,0),
        bits_list=[Si],
        sections=[("Preambulo $S_I$", len(Si))],
        colors=["darkgreen"],
        ylabel="Canal $I$"
    ).plot()
    
    BitsPlot(
        fig_preamble, grid_preamble, (1,0),
        bits_list=[Sq],
        sections=[("Preambulo $S_Q$", len(Sq))],
        colors=["navy"],
        xlabel="Index de Bit", 
        ylabel="Canal $Q$"
    ).plot()
    
    fig_preamble.tight_layout()
    save_figure(fig_preamble, "example_preamble.pdf")
        
    


# """
# Implementação de um canal para agregação de multiplos sinais.

# Autor: Arthur Cadore
# Data: 16-08-2025
# """

import numpy as np

from .transmitter import Transmitter
from .datagram import Datagram
from .plotter import save_figure, create_figure, TimePlot, FrequencyPlot
from .noise import Noise, NoiseEBN0


class Channel:
    def __init__(self, fs=128_000, duration=1, noise_mode="snr", noise_db=20, seed=10):
        r"""
        Implementação de um canal para agregação de multiplos sinais.
        
        Args:
            fs (int): taxa de amostragem do sinal.
            duration (int): duração do canal em segundos.
            noise_mode (str): modo de ruído ('snr' ou 'ebn0').
            noise_db (int): nível de ruído em dB.
            seed (int): semente para geração de números aleatórios. 
    
        Returns:
            Channel: objeto do canal.
    
        Raises:
            ValueError: se o modo de ruído for inválido.
        """
        self.fs = fs
        self.channel = np.zeros(int(fs * duration))
        self.t = np.arange(0, duration, 1/fs)

        noise_map = {
            "ebn0": 0,
            "snr": 1
        }

        noise_mode = noise_mode.lower()
        if noise_mode not in noise_map:
            raise ValueError("Modo de ruído inválido. Use 'EBN0', 'SNR'.")

        self.noise_mode = noise_map[noise_mode]
        self.noise_db = noise_db
        self.seed = seed

    def add_signal(self, signal, position_factor=0.5):
        r"""
        Adiciona um sinal ao canal em uma posição relativa.

        Args:
            signal (np.ndarray): vetor de amostras do sinal a inserir.
            position_factor (float): fator de posição entre [0, 1] (0 = início do canal, 1 = final).

        Raises:
            ValueError: se position_factor não estiver entre [0, 1].
        
        Example: 
            ![pageplot](assets/example_channel_time_subchannels.svg)
        """
        if not 0 <= position_factor <= 1:
            raise ValueError("position_factor deve estar entre 0 e 1.")

        chan_len = len(self.channel)
        sig_len = len(signal)

        # Calcula posição inicial no vetor do canal
        start_idx = int(round(position_factor * (chan_len - sig_len)))
        if start_idx < 0:
            start_idx = 0
        if start_idx + sig_len > chan_len:
            sig_len = chan_len - start_idx
            signal = signal[:sig_len]  # corta se não couber

        # Insere (soma) o sinal no canal
        self.channel[start_idx:start_idx + sig_len] += signal

    def add_noise(self):
        r"""
        Adiciona ruído ao canal.

        Example: 
            ![pageplot](assets/example_channel_time_channel.svg)
        """
        if self.noise_mode == 0:
            noise = NoiseEBN0(ebn0_db=self.noise_db, seed=self.seed)
        elif self.noise_mode == 1:
            noise = Noise(snr=self.noise_db, seed=self.seed)
        
        self.channel = noise.add_noise(self.channel)
    

if __name__ == "__main__":

    # Cria transmissor e gera o vetor de sinal.
    tx = Transmitter()
    datagram = Datagram(pcdnum=1234, numblocks=1, seed=10)
    t, s = tx.transmit(datagram)
    
    # Cria o canal
    canal1 = Channel(fs=tx.fs, duration=1, noise_mode="snr", noise_db=20, seed=10)
    canal2 = Channel(fs=tx.fs, duration=1, noise_mode="snr", noise_db=20, seed=10)
    canal3 = Channel(fs=tx.fs, duration=1, noise_mode="snr", noise_db=20, seed=10)
    
    # coloca o sinal no meio do canal
    canal1.add_signal(s, position_factor=0.1)
    canal2.add_signal(s, position_factor=0.5)
    canal3.add_signal(s, position_factor=0.9)

    fig_time, grid = create_figure(4, 1, figsize=(16, 9))

    TimePlot(
        fig_time, grid, (0, 0),
        t=np.arange(0, len(s)/tx.fs, 1/tx.fs),
        signals=[s],
        labels=["$s(t)$"],
        title="Sinal Modulado $s(t)$",
        colors=["darkred"],
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()    
    
    TimePlot(
        fig_time, grid, (1, 0),
        t=canal1.t,
        signals=[canal1.channel],
        labels=["$s(t)$"], 
        title="Sinal Modulado - Canal 1",
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    TimePlot(
        fig_time, grid, (2, 0),
        t=canal2.t,
        signals=[canal2.channel],
        labels=["$s(t)$"], 
        title="Sinal Modulado - Canal 2",
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    TimePlot(
        fig_time, grid, (3, 0),
        t=canal3.t,
        signals=[canal3.channel],
        labels=["$s(t)$"], 
        title="Sinal Modulado - Canal 3",
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    fig_time.tight_layout()
    save_figure(fig_time, "example_channel_time_subchannels.pdf")
    
    canalT = canal1.channel + canal2.channel + canal3.channel

    canalT_NoiseEBN0 = NoiseEBN0(ebn0_db=20, seed=10).add_noise(canalT)

    fig_time, grid = create_figure(2, 1, figsize=(16, 9))
    
    TimePlot(
        fig_time, grid, (0, 0),
        t=np.arange(0, len(canalT)/tx.fs, 1/tx.fs),
        signals=[canalT],
        labels=["$s(t)$"], 
        title="Sinal Modulado - Canal Total",
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    TimePlot(
        fig_time, grid, (1, 0),
        t=np.arange(0, len(canalT_NoiseEBN0)/tx.fs, 1/tx.fs),
        signals=[canalT_NoiseEBN0],
        labels=["$s(t)$"], 
        title="Sinal Modulado - Canal Total + Ruido",
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    fig_time.tight_layout()
    save_figure(fig_time, "example_channel_time_channel.pdf")
    
        
    


# """
# Implementação de simulação para curva BER vs Eb/N0. 

# Autor: Arthur Cadore
# Data: 8-09-2025
# """

import numpy as np
import concurrent.futures
from scipy.special import erfc
from tqdm import tqdm

from .datagram import Datagram
from .transmitter import Transmitter
from .receiver import Receiver
from .noise import NoiseEBN0
from .data import ExportData, ImportData
from .plotter import create_figure, save_figure, BersnrPlot

def interpolate(positions, ref_points, ref_values):
    r"""
    Define o número de repetições em função do $Eb/N0$, usando interpolação linear entre pontos de referência, dada pela expressão abaixo.

    $$
    r = r_{i} + \frac{(EBN0 - EBN0_{i})}{(EBN0_{i+1} - EBN0_{i})} \cdot (r_{i+1} - r_{i})
    $$

    Onde:
        - $r$: Número de repetições.
        - $EBN0$: Relação $Eb/N_0$ em decibéis.
        - $r_i$ e $r_{i+1}$: Número de repetições nos pontos de referência próximos.
        - $EBN0_i$ e $EBN0_{i+1}$: Relações $Eb/N_0$ nos pontos de referência próximos.

    Args: 
        positions (int): O número total de pontos a serem gerados.
        ref_points (array-like): Pontos de referência. 
        ref_values (array-like): Valores correspondentes aos pontos de referência.
    
    Returns:
        interpolated_values (np.ndarray): Vetor de valores interpolados, arredondados para inteiros.
    """
    # Garante que as entradas são arrays numpy
    ref_points = np.array(ref_points)
    ref_values = np.array(ref_values)
    
    # Realiza a interpolação linear usando np.interp
    interpolated_values = np.interp(np.linspace(ref_points[0], ref_points[-1], positions), ref_points, ref_values)
    
    # Arredonda os valores e converte para inteiros
    interpolated_values = np.round(interpolated_values).astype(int)
    
    return interpolated_values

class BERSNR_ARGOS: 
    def __init__(self, EbN0_values=np.arange(0, 10, 1), num_workers=56, numblocks=8, max_repetitions=2000, error_values=None):
        r"""
        Implementa a simulação de BER vs Eb/N0 para o padrão ARGOS-3.

        Args:
            EbN0_values (array-like): Valores de Eb/N0 para os quais a simulação será realizada.
            num_workers (int): Número de threads para paralelização.
            numblocks (int): Número de blocos de dados para cada datagrama.
            max_repetitions (int): Número máximo de repetições para cada valor de Eb/N0.
            error_values (array-like): Número máximo de erros para cada valor de Eb/N0.
        
        Raises:
            ValueError: Se o número de erros não for o mesmo que o número de valores de Eb/N0.

        Example: 
            ![pageplot](assets/ber_vs_ebn0.svg)
        """
        if len(error_values) != len(EbN0_values):
            raise ValueError("error_values deve ter o mesmo tamanho que EbN0_values")

        # Parâmetros fixos do sistema
        self.fs = 128_000
        self.Rb = 400
        self.fc = 4000

        # Parâmetros variáveis
        self.EbN0_values = EbN0_values
        self.num_workers = num_workers
        self.numblocks = numblocks
        self.max_repetitions = max_repetitions
        self.error_values = error_values

        # Cadeia de TX
        self.datagramTX = Datagram(pcdnum=1234, numblocks=numblocks, seed=10)
        self.bitsTX = self.datagramTX.streambits
        self.bitsSent = len(self.bitsTX)

        # Gerando sinal fixo s(t)
        self.t, self.s = Transmitter(fc=self.fc, output_print=False, output_plot=False, fs=self.fs, Rb=self.Rb).transmit(datagram=self.datagramTX) 

        # Cadeia de RX
        self.receiver = Receiver(fc=self.fc, output_print=False, output_plot=False, fs=self.fs, Rb=self.Rb) 

    def simulate(self, ebn0_db):
        # Adicionando ruído ao sinal
        add_noise = NoiseEBN0(ebn0_db, fs=self.fs, Rb=self.Rb, seed=10)
        s_noisy = add_noise.add_noise(self.s)
        
        # Recebendo bits
        bitsRX = self.receiver.receive(s_noisy)

        # Contando erros entre bitsTX e bitsRX
        num_errors = sum(1 for tx, rx in zip(self.bitsTX, bitsRX) if tx != rx)
        return num_errors

    def run(self):
        r"""
        Executa a simulação de BER vs Eb/N0 para o padrão ARGOS-3.

        Returns:
            ber_results (list): Lista de tuplas (Eb/N0, BER) para cada valor de Eb/N0.
        """
        ber_results = []

        # Paralelizar as simulações para cada Eb/N0 usando Pool de Workers
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            # monitorar o progresso da simulação em cada iteração de Eb/N0
            for ebn0_db in self.EbN0_values:
                total_errors = 0
                repetitions = 0

                with tqdm(total=self.max_repetitions, desc=f"Simulando Eb/N0 = {ebn0_db} dB", ncols=100) as pbar:
                    while repetitions < self.max_repetitions and total_errors < self.error_values[int(ebn0_db)]:

                        # Cria múltiplas simulações (uma para cada worker)
                        futures = [executor.submit(self.simulate, ebn0_db) for _ in range(self.num_workers)]  

                        for future in futures:
                            # Aguarda a conclusão da tarefa
                            num_errors = future.result() 
                            total_errors += num_errors
                            repetitions += 1
                            pbar.update(1)  

                        # Se atingiu o limite de erros, interrompe a simulação
                        if total_errors >= self.error_values[int(ebn0_db)]:
                            break

                # Calcula o número total de bits transmitidos (Repetições * Bits do datagrama)
                total_bits_transmitted = repetitions * self.bitsSent

                # Calcula a BER
                if total_bits_transmitted > 0:
                    ber = (total_errors + 1) / (total_bits_transmitted + 1)
                else:
                    ber = 0

                # Status da simulação
                print(f"[ARGOS-3] Eb/N0={ebn0_db} dB -> Bits={total_bits_transmitted}, Erros={total_errors}, BER={ber}")

                # Armazena a tupla (Eb/N0, BER) na lista
                ber_results.append((ebn0_db, ber))

        return ber_results

class BERSNR_QPSK:
    def __init__(self, EbN0_values=np.arange(0, 10, 1), num_workers=8, num_bits=10_000, max_repetitions=2000, error_values=None):
        r"""
        Implementa a simulação de BER vs Eb/N0 para o padrão QPSK.

        Args:
            EbN0_values (array-like): Valores de Eb/N0 para os quais a simulação será realizada.
            num_workers (int): Número de threads para paralelização.
            num_bits (int): Número de bits para cada simulação.
            max_repetitions (int): Número máximo de repetições para cada valor de Eb/N0.
            error_values (array-like): Número máximo de erros para cada valor de Eb/N0.
        
        Raises:
            ValueError: Se o número de erros não for o mesmo que o número de valores de Eb/N0.
        """

        if error_values is None or len(error_values) != len(EbN0_values):
            raise ValueError("error_values deve ter o mesmo tamanho que EbN0_values")

        # Parâmetros variáveis
        self.EbN0_values = EbN0_values
        self.num_workers = num_workers
        self.num_bits = num_bits
        self.max_repetitions = max_repetitions
        self.error_values = error_values

    @staticmethod
    def simulate_qpsk(ebn0_db, num_bits=1000, bits_por_simbolo=2, rng=10):
        # Seed do gerador de números aleatórios
        rng = np.random.default_rng(rng)

        # Geração dos bits (I e Q independentes)
        bI = rng.integers(0, 2, size=(num_bits,))
        bQ = rng.integers(0, 2, size=(num_bits,))

        # Mapeamento QPSK
        I = (2*bI - 1) / np.sqrt(2)
        Q = (2*bQ - 1) / np.sqrt(2)

        # Sinal complexo
        signal = I + 1j*Q

        # Cálculo do Eb/N0
        ebn0_lin = 10 ** (ebn0_db / 10)
        signal_power = np.mean(np.abs(signal)**2)
        bit_energy = signal_power / bits_por_simbolo
        noise_density = bit_energy / ebn0_lin
        variance = noise_density / 2
        sigma = np.sqrt(variance)

        # Canal AWGN
        noise = rng.normal(0.0, sigma, size=signal.shape) + 1j * rng.normal(0.0, sigma, size=signal.shape)
        r = signal + noise

        # Demodulação
        bI_dec = (r.real >= 0).astype(int)
        bQ_dec = (r.imag >= 0).astype(int)

        # Contagem de erros
        erros = np.count_nonzero(bI_dec != bI) + np.count_nonzero(bQ_dec != bQ)
        ber = erros / (2 * num_bits)
        return ber

    def run(self):
        r"""
        Executa a simulação de BER vs Eb/N0 para o padrão QPSK.

        Returns:
            ber_results (list): Lista de tuplas (Eb/N0, BER) para cada valor de Eb/N0.
        """
        ber_results = []

        # Paralelizar as simulações para cada Eb/N0 usando Pool de Workers
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            for ebn0_db in self.EbN0_values:
                total_errors = 0
                repetitions = 0
                total_bits = 0

                # Monitorar o progresso da simulação em cada iteração de Eb/N0
                with tqdm(total=self.max_repetitions, desc=f"QPSK Eb/N0 = {ebn0_db} dB", ncols=100) as pbar:

                    # Executar as simulações para cada worker
                    while repetitions < self.max_repetitions and total_errors < self.error_values[int(ebn0_db)]:
                        futures = [executor.submit(self.simulate_qpsk, ebn0_db, num_bits=self.num_bits, rng=np.random.default_rng())
                                   for _ in range(self.num_workers)]

                        # Aguardar a conclusão das simulações para cada worker
                        for future in concurrent.futures.as_completed(futures):
                            ber = future.result()

                            # Contagem de erros (2 bits por símbolo)
                            errors = int(ber * self.num_bits * 2) 
                            total_errors += errors
                            total_bits += self.num_bits * 2
                            repetitions += 1
                            pbar.update(1)

                        # Se atingiu o limite de erros, interrompe a simulação
                        if total_errors >= self.error_values[int(ebn0_db)]:
                            break

                # Calcula a BER final
                if total_bits > 0:
                    ber_final = (total_errors + 1) / (total_bits + 1)
                else:
                    ber_final = 0

                # Status da simulação
                print(f"[QPSK] Eb/N0={ebn0_db} dB -> Bits={total_bits}, Erros={total_errors}, BER={ber_final}")

                # Armazena a tupla (Eb/N0, BER) na lista
                ber_results.append((ebn0_db, ber_final))

        return ber_results

    def teorical_qpsk(self):
        r"""
        Calcula a curva teórica de $BER$ vs $Eb/N_0$ para QPSK, segundo a expressão abaixo.

        $$
        P_b(x) = Q \left(x\right) \mapsto P_b(x) = Q\left(\sqrt{2 \cdot \frac{E_b}{N_0}}\right)
        $$

        Sendo:
            - $P_b(x)$: Probabilidade de erro. 
            - $Q(x)$: Função de erro complementar.
            - $x$: Argumento da função $Q(x)$.
            - $E_b$: Energia por bit.
            - $N_0$: Potência do ruído. 

        Returns:
            ber_teorico (np.ndarray): Vetor de valores de BER teórica para cada Eb/N0 da classe.
        """

        ebn0_lin = 10 ** (self.EbN0_values / 10)
        
        # argumento da função Q(x)
        x = np.sqrt(2 * ebn0_lin)

        # calculo da função Q(x)
        Qx = 0.5 * erfc(x / np.sqrt(2))
        return Qx



if __name__ == "__main__":

    # Define os valores de Eb/N0 para a simulação
    EbN0_vec = np.arange(0, 9.5, 0.5)

    ref_values = [10000, 5000, 800, 200]
    ref_points = [0, 3, 6, 12]
    error_values = interpolate(len(EbN0_vec), ref_points, ref_values)

    # Imprime os valores de erro máximo para cada Eb/N0
    for ebn0, error in zip(EbN0_vec, error_values):
        print(f"Eb/N0 = {ebn0} dB: {error} erros")

    ### ARGOS-3
    reps = 1048576
    print(f"[ARGOS-3] Maximo de bits transmitidos por Eb/N0: {reps}")
    bersnr_argos = BERSNR_ARGOS(EbN0_values=EbN0_vec, error_values=error_values, num_workers=64, numblocks=1, max_repetitions=reps)

    ### QPSK
    bersnr_qpsk = BERSNR_QPSK(EbN0_values=EbN0_vec, error_values=error_values, num_workers=56, num_bits=50_000, max_repetitions=5000)

    # Simulação
    # ###############################################

    results = bersnr_argos.run()
    ExportData(results, "bersnr_argos").save()

    results_qpsk = bersnr_qpsk.run()
    ExportData(results_qpsk, "bersnr_qpsk").save()
    
    # PLOT
    # ###############################################

    # extrair os valores de Eb/N0 e BER
    bersnr_argos = ImportData("bersnr_argos").load()
    ber_values_argos = bersnr_argos[:, 1]

    print(ber_values_argos)

    # QPSK Teorico
    bersnr_qpsk_teorico = bersnr_qpsk.teorical_qpsk()
    
    # extrair os valores de Eb/N0 e BER
    bersnr_qpsk = ImportData("bersnr_qpsk").load()
    ber_values_qpsk = bersnr_qpsk[:, 1]

    print(ber_values_qpsk)
    print(bersnr_qpsk_teorico)


    # extrair os valores de Eb/N0 e BER teórico
    fig, grid = create_figure(1, 1)
    BersnrPlot(fig, grid, 0,
               EbN0=EbN0_vec,
               ber_curves=[ber_values_argos, ber_values_qpsk, bersnr_qpsk_teorico],
               labels=["ARGOS-3", "QPSK Simulado", "QPSK Ideal"],
               linestyles=["-", "-", ":"],
               markers=["o", "s", "x"],
               title="Curva BER vs Eb/N0",
               ylim=(1e-5, 1),
               xlim=(-1, 10)
    ).plot()
    
    save_figure(fig, "ber_vs_ebn0.pdf")
